import astropy.constants as c
import astropy.units as u
import pandas as pd
import numpy as np
import os
from scipy import optimize 
import PyMieScatt as ps

from .root_functions import advdiff, vfall,vfall_find_root,qvs_below_model, find_cond_t, solve_force_balance
from .calc_mie import fort_mie_calc, calc_new_mieff
from . import gas_properties
from . import pvaps
from .justplotit import plot_format, find_nearest_1d

from .direct_mmr_solver import direct_solver


def compute(atmo, directory = None, as_dict = True, og_solver = True, 
    direct_tol=1e-15, refine_TP = True, og_vfall=True, analytical_rg = True, do_virtual=True):

    """
    Top level program to run eddysed. Requires running `Atmosphere` class 
    before running this. 
    
    Parameters
    ----------
    atmo : class 
        `Atmosphere` class 
    directory : str, optional 
        Directory string that describes where refrind files are 
    as_dict : bool, optional 
        Default = False. Option to view full output as dictionary
    og_solver : bool, optional
        Default=True. BETA. Contact developers before changing to False.
         Option to change mmr solver (True = original eddysed, False = new direct solver)
    direct_tol : float , optional
        Only used if og_solver =False. Default = True. 
        Tolerance for direct solver
    refine_TP : bool, optional
        Only used if og_solver =False. 
        Option to refine temperature-pressure profile for direct solver 
    og_vfall : bool, optional
        Option to use original A&M or new Khan-Richardson method for finding vfall
    analytical_rg : bool, optional
        Only used if og_solver =False. 
        Option to use analytical expression for rg, or alternatively deduce rg from calculation
        Calculation option will be most useful for future 
        inclusions of alternative particle size distributions
    do_virtual : bool 
        If the user adds an upper bound pressure that is too low. There are cases where a cloud wants to 
        form off the grid towards higher pressures. This enables that. 

    Returns 
    -------
    opd, w0, g0
        Extinction per layer, single scattering abledo, asymmetry parameter, 
        All are ndarrays that are nlayer by nwave
    dict 
        Dictionary output that contains full output. See tutorials for explanation of all output.
    """
    mmw = atmo.mmw
    mh = atmo.mh
    condensibles = atmo.condensibles

    ngas = len(condensibles)

    gas_mw = np.zeros(ngas)
    gas_mmr = np.zeros(ngas)
    rho_p = np.zeros(ngas)

    # scale-height for fsed taken at Teff (default: temp at 1bar) 
    H = atmo.r_atmos * atmo.Teff / atmo.g
    
    
    #### First we need to either grab or compute Mie coefficients #### 
    for i, igas in zip(range(ngas),condensibles) : 

        #Get gas properties including gas mean molecular weight,
        #gas mixing ratio, and the density
        run_gas = getattr(gas_properties, igas)
        gas_mw[i], gas_mmr[i], rho_p[i] = run_gas(mmw, mh=mh, gas_mmr=atmo.gas_mmr)

        #Get mie files that are already saved in 
        #directory
        #eventually we will replace this with nice database 
        qext_gas, qscat_gas, cos_qscat_gas, nwave, radius,wave_in = get_mie(igas,directory)

        if i==0: 
            nradii = len(radius)
            rmin = np.min(radius)
            radius, rup, dr = get_r_grid(rmin, n_radii=nradii)
            qext = np.zeros((nwave,nradii,ngas))
            qscat = np.zeros((nwave,nradii,ngas))
            cos_qscat = np.zeros((nwave,nradii,ngas))

        #add to master matrix that contains the per gas Mie stuff
        qext[:,:,i], qscat[:,:,i], cos_qscat[:,:,i] = qext_gas, qscat_gas, cos_qscat_gas

    #Next, calculate size and concentration
    #of condensates in balance between eddy diffusion and sedimentation

    #qc = condensate mixing ratio, qt = condensate+gas mr, rg = mean radius,
    #reff = droplet eff radius, ndz = column dens of condensate, 
    #qc_path = vertical path of condensate

    #   run original eddysed code
    if og_solver:
        #here atmo.param describes the parameterization used for the variable fsed methodology
        if atmo.param is 'exp': 
            #the formalism of this is detailed in Rooney et al. 2021
            atmo.b = 6 * atmo.b * H # using constant scale-height in fsed
            fsed_in = (atmo.fsed-atmo.eps) 
        elif atmo.param is 'const':
            fsed_in = atmo.fsed
        qc, qt, rg, reff, ndz, qc_path, mixl, z_cld = eddysed(atmo.t_level, atmo.p_level, atmo.t_layer, atmo.p_layer, 
                                             condensibles, gas_mw, gas_mmr, rho_p , mmw, 
                                             atmo.g, atmo.kz, atmo.mixl, 
                                             fsed_in,
                                             atmo.b, atmo.eps, atmo.scale_h, atmo.z_top, atmo.z_alpha, min(atmo.z), atmo.param,
                                             mh, atmo.sig, rmin, nradii,
                                             atmo.d_molecule,atmo.eps_k,atmo.c_p_factor,
                                             og_vfall, supsat=atmo.supsat,verbose=atmo.verbose,do_virtual=do_virtual)
        pres_out = atmo.p_layer
        temp_out = atmo.t_layer
        z_out = atmo.z

    
    #   run new, direct solver
    else:
        qc, qt, rg, reff, ndz, qc_path, pres_out, temp_out, z_out,mixl = direct_solver(atmo.t_layer, atmo.p_layer,
                                             condensibles, gas_mw, gas_mmr, rho_p , mmw, 
                                             atmo.g, atmo.kz, atmo.fsed, mh,atmo.sig, rmin, nradii, 
                                             atmo.d_molecule,atmo.eps_k,atmo.c_p_factor,
                                             direct_tol, refine_TP, og_vfall, analytical_rg)

            
    #Finally, calculate spectrally-resolved profiles of optical depth, single-scattering
    #albedo, and asymmetry parameter.    
    opd, w0, g0, opd_gas = calc_optics(nwave, qc, qt, rg, reff, ndz,radius,
                                       dr,qext, qscat,cos_qscat,atmo.sig, rmin, nradii)

    if as_dict:
        if atmo.param is 'exp':
            fsed_out = fsed_in * np.exp((atmo.z - atmo.z_alpha) / atmo.b ) + atmo.eps
        else: 
            fsed_out = fsed_in 
        return create_dict(qc, qt, rg, reff, ndz,opd, w0, g0, 
                           opd_gas,wave_in, pres_out, temp_out, condensibles,
                           mh,mmw, fsed_out, atmo.sig, nradii,rmin, z_out, atmo.dz_layer, 
                           mixl, atmo.kz, atmo.scale_h, z_cld) 
    else:
        return opd, w0, g0

def create_dict(qc, qt, rg, reff, ndz,opd, w0, g0, opd_gas,wave,pressure,temperature, gas_names,
    mh,mmw,fsed,sig,nrad,rmin,z, dz_layer, mixl, kz, scale_h, z_cld):
    return {
        "pressure":pressure/1e6, 
        "pressure_unit":'bar',
        "temperature":temperature,
        "temperature_unit":'kelvin',
        "wave":wave[:,0],
        "wave_unit":'micron',
        "condensate_mmr":qc,
        "cond_plus_gas_mmr":qt,
        "mean_particle_r":rg*1e4,
        "droplet_eff_r":reff*1e4, 
        "r_units":'micron',
        "column_density":ndz,
        "column_density_unit":'#/cm^2',
        "opd_per_layer":opd, 
        "single_scattering" : w0, 
        "asymmetry": g0, 
        "opd_by_gas": opd_gas,
        "condensibles":gas_names,
        #"scalar_inputs": {'mh':mh, 'mmw':mmw,'fsed':fsed, 'sig':sig,'nrad':nrad,'rmin':rmin},
        "scalar_inputs": {'mh':mh, 'mmw':mmw,'sig':sig,'nrad':nrad,'rmin':rmin},
        "fsed": fsed,
        "altitude":z,
        "layer_thickness":dz_layer,
        "z_unit":'cm',
        'mixing_length':mixl, 
        'mixing_length_unit':'cm',
        'kz':kz, 
        'kz_unit':'cm^2/s',
        'scale_height':scale_h,
        'cloud_deck':z_cld
    }

def calc_optics(nwave, qc, qt, rg, reff, ndz,radius,dr,qext, qscat,cos_qscat,sig, rmin, nrad):
    """
    Calculate spectrally-resolved profiles of optical depth, single-scattering
    albedo, and asymmetry parameter.

    Parameters
    ----------
    nwave : int 
        Number of wave points 
    qc : ndarray
        Condensate mixing ratio 
    qt : ndarray 
        Gas + condensate mixing ratio 
    rg : ndarray 
        Geometric mean radius of condensate 
    reff : ndarray
        Effective (area-weighted) radius of condensate (cm)
    ndz : ndarray
        Column density of particle concentration in layer (#/cm^2)
    radius : ndarray
        Radius bin centers (cm)
    dr : ndarray
        Width of radius bins (cm)
    qscat : ndarray
        Scattering efficiency
    qext : ndarray
        Extinction efficiency
    cos_qscat : ndarray
        qscat-weighted <cos (scattering angle)>
    sig : float 
        Width of the log normal particle distribution


    Returns
    -------
    opd : ndarray 
        extinction optical depth due to all condensates in layer
    w0 : ndarray 
        single scattering albedo
    g0 : ndarray 
        asymmetry parameter = Q_scat wtd avg of <cos theta>
    opd_gas : ndarray
        cumulative (from top) opd by condensing vapor as geometric conservative scatterers
    """

    PI=np.pi
    nz = qc.shape[0]
    ngas = qc.shape[1]
    nrad = len(radius)

    opd_layer = np.zeros((nz, ngas))
    scat_gas = np.zeros((nz,nwave,ngas))
    ext_gas = np.zeros((nz,nwave,ngas))
    cqs_gas = np.zeros((nz,nwave,ngas))
    opd = np.zeros((nz,nwave))
    opd_gas = np.zeros((nz,ngas))
    w0 = np.zeros((nz,nwave))
    g0 = np.zeros((nz,nwave))

    for iz in range(nz):
        for igas in range(ngas):
            # Optical depth for conservative geometric scatterers 
            if ndz[iz,igas] > 0:

                if np.log10(rg[iz,igas]) < np.log10(rmin)+0.75*sig:
                    raise Exception ('There has been a calculated particle radii of {0}cm for the {1}th gas at the {2}th grid point. The minimum radius from the Mie grid is {3}cm, and youve requested a lognormal distribution of {4}. Therefore it is not possible to accurately compute the optical properties.'.format(str(rg[iz,igas]),str(igas),str(iz), str(rmin),str(sig)))

                r2 = rg[iz,igas]**2 * np.exp( 2*np.log( sig)**2 )
                opd_layer[iz,igas] = 2.*PI*r2*ndz[iz,igas]

                #  Calculate normalization factor (forces lognormal sum = 1.0)
                rsig = sig
                norm = 0.
                for irad in range(nrad):
                    rr = radius[irad]
                    arg1 = dr[irad] / ( np.sqrt(2.*PI)*rr*np.log(rsig) )
                    arg2 = -np.log( rr/rg[iz,igas] )**2 / ( 2*np.log(rsig)**2 )
                    norm = norm + arg1*np.exp( arg2 )
                    #print (rr, rg[iz,igas],rsig,arg1,arg2)

                # normalization
                norm = ndz[iz,igas] / norm

                for irad in range(nrad):
                    rr = radius[irad]
                    arg1 = dr[irad] / ( np.sqrt(2.*PI)*np.log(rsig) )
                    arg2 = -np.log( rr/rg[iz,igas] )**2 / ( 2*np.log(rsig)**2 )
                    pir2ndz = norm*PI*rr*arg1*np.exp( arg2 )         
                    for iwave in range(nwave): 
                        scat_gas[iz,iwave,igas] = scat_gas[iz,iwave,igas]+qscat[iwave,irad,igas]*pir2ndz
                        ext_gas[iz,iwave,igas] = ext_gas[iz,iwave,igas]+qext[iwave,irad,igas]*pir2ndz
                        cqs_gas[iz,iwave,igas] = cqs_gas[iz,iwave,igas]+cos_qscat[iwave,irad,igas]*pir2ndz

                    #TO DO ADD IN CLOUD SUBLAYER KLUGE LATER 

    #Sum over gases and compute spectral optical depth profile etc
    for iz in range(nz):
        for iwave in range(nwave): 
            opd_scat = 0.
            opd_ext = 0.
            cos_qs = 0.
            for igas in range(ngas):
                opd_scat = opd_scat + scat_gas[iz,iwave,igas]
                opd_ext = opd_ext + ext_gas[iz,iwave,igas]
                cos_qs = cos_qs + cqs_gas[iz,iwave,igas]

                if( opd_scat > 0. ):
                    opd[iz,iwave] = opd_ext
                    w0[iz,iwave] = opd_scat / opd_ext
                    #if w0[iz,iwave]>1: 
                    #    w0[iz,iwave]=1.
                    g0[iz,iwave] = cos_qs / opd_scat
                    
    #cumulative optical depths for conservative geometric scatterers
    opd_tot = 0.

    for igas in range(ngas):
        opd_gas[0,igas] = opd_layer[0,igas]

        for iz in range(1,nz):
            opd_gas[iz,igas] = opd_gas[iz-1,igas] + opd_layer[iz,igas]

    return opd, w0, g0, opd_gas

def eddysed(t_top, p_top,t_mid, p_mid, condensibles, 
    gas_mw, gas_mmr,rho_p,mw_atmos,gravity, kz,mixl,
    fsed, b, eps, scale_h, z_top, z_alpha, z_min, param,
    mh,sig, rmin, nrad,d_molecule,eps_k,c_p_factor,
    og_vfall=True,do_virtual=True, supsat=0, verbose=True):
    """
    Given an atmosphere and condensates, calculate size and concentration
    of condensates in balance between eddy diffusion and sedimentation.

    Parameters
    ----------
    t_top : ndarray
        Temperature at each layer (K)
    p_top : ndarray
        Pressure at each layer (dyn/cm^2)
    t_mid : ndarray
        Temperature at each midpoint (K)
    p_mid : ndarray 
        Pressure at each midpoint (dyn/cm^2)
    condensibles : ndarray or list of str
        List or array of condensible gas names
    gas_mw : ndarray
        Array of gas mean molecular weight from `gas_properties`
    gas_mmr : ndarray 
        Array of gas mixing ratio from `gas_properties`
    rho_p : float 
        density of condensed vapor (g/cm^3)
    mw_atmos : float 
        Mean molecular weight of the atmosphere
    gravity : float 
        Gravity of planet cgs
    kz : float or ndarray
        Kzz in cgs, either float or ndarray depending of whether or not 
        it is set as input
    fsed : float 
        Sedimentation efficiency coefficient, unitless
    b : float
        Denominator of exponential in sedimentation efficiency  (if param is 'exp')
    eps: float
        Minimum value of fsed function (if param=exp)
    scale_h : float 
        Scale height of the atmosphere
    z_top : float
        Altitude at each layer
    z_alpha : float
        Altitude at which fsed=alpha for variable fsed calculation
    param : str
        fsed parameterisation
        'const' (constant), 'exp' (exponential density derivation)
    mh : float 
        Atmospheric metallicity in NON log units (e.g. 1 for 1x solar)
    sig : float 
        Width of the log normal particle distribution
    d_molecule : float 
        diameter of atmospheric molecule (cm) (Rosner, 2000)
        (3.711e-8 for air, 3.798e-8 for N2, 2.827e-8 for H2)
        Set in Atmosphere constants 
    eps_k : float 
        Depth of the Lennard-Jones potential well for the atmosphere 
        Used in the viscocity calculation (units are K) (Rosner, 2000)
    c_p_factor : float 
        specific heat of atmosphere (erg/K/g) . Usually 7/2 for ideal gas
        diatomic molecules (e.g. H2, N2). Technically does slowly rise with 
        increasing temperature
    og_vfall : bool , optional
        optional, default = True. True does the original fall velocity calculation. 
        False does the updated one which runs a tad slower but is more consistent.
        The main effect of turning on False is particle sizes in the upper atmosphere 
        that are slightly bigger.
    do_virtual : bool,optional 
        optional, Default = True which adds a virtual layer if the 
        species condenses below the model domain.
    supsat : float, optional
        Default = 0 , Saturation factor (after condensation)

    Returns
    -------
    qc : ndarray 
        condenstate mixing ratio (g/g)
    qt : ndarray 
        gas + condensate mixing ratio (g/g)
    rg : ndarray
        geometric mean radius of condensate  cm 
    reff : ndarray
        droplet effective radius (second moment of size distrib, cm)
    ndz : ndarray 
        number column density of condensate (cm^-3)
    qc_path : ndarray 
        vertical path of condensate 
    """
    #default for everything is false, will fill in as True as we go


    did_gas_condense = [False for i in condensibles]
    t_bot = t_top[-1]
    p_bot = p_top[-1]
    z_bot = z_top[-1]
    ngas =  len(condensibles)
    nz = len(t_mid)
    qc = np.zeros((nz,ngas))
    qt  = np.zeros((nz, ngas))
    rg = np.zeros((nz, ngas))
    reff = np.zeros((nz, ngas))
    ndz = np.zeros((nz, ngas))
    fsed_layer = np.zeros((nz,ngas))
    qc_path = np.zeros(ngas)
    z_cld_out = np.zeros(ngas)

    for i, igas in zip(range(ngas), condensibles):

        q_below = gas_mmr[i]

        #include decrease in condensate mixing ratio below model domain
        if do_virtual: 

            qvs_factor = (supsat+1)*gas_mw[i]/mw_atmos
            get_pvap = getattr(pvaps, igas)
            if igas == 'Mg2SiO4':
                pvap = get_pvap(t_bot, p_bot, mh=mh)
            else:
                pvap = get_pvap(t_bot, mh=mh)

            qvs = qvs_factor*pvap/p_bot   
            if qvs <= q_below :   

                #find the pressure at cloud base 
                #   parameters for finding root 
                p_lo = p_bot
                p_hi = p_bot * 1e3

                #temperature gradient 
                dtdlnp = ( t_top[-2] - t_bot ) / np.log( p_bot/p_top[-2] )

                #   load parameters into qvs_below common block
                qv_dtdlnp = dtdlnp
                qv_p = p_bot
                qv_t = t_bot
                qv_gas_name = igas
                qv_factor = qvs_factor

                try:

                    p_base = optimize.root_scalar(qvs_below_model, 
                                bracket=[p_lo, p_hi], method='brentq', 
                                args=(qv_dtdlnp,qv_p, qv_t,qv_factor ,qv_gas_name,mh,q_below)
                                )#, xtol = 1e-20)

                    if verbose: print('Virtual Cloud Found: '+ qv_gas_name)
                    root_was_found = True
                except ValueError: 
                    root_was_found = False

                if root_was_found:
                    #Yes, the gas did condense (below the grid)
                    did_gas_condense[i] = True

                    p_base = p_base.root 
                    t_base = t_bot + np.log( p_bot/p_base )*dtdlnp
                    z_base = z_bot + scale_h[-1] * np.log( p_bot/p_base ) 
                    
                    #   Calculate temperature and pressure below bottom layer
                    #   by adding a virtual layer 

                    p_layer_virtual = 0.5*( p_bot + p_base )
                    t_layer_virtual = t_bot + np.log10( p_bot/p_layer_virtual )*dtdlnp

                    #we just need to overwrite 
                    #q_below from this output for the next routine
                    qc_v, qt_v, rg_v, reff_v,ndz_v,q_below, z_cld, fsed_layer_v = layer( igas, rho_p[i], 
                        #t,p layers, then t.p levels below and above
                        t_layer_virtual, p_layer_virtual, t_bot,t_base, p_bot, p_base,
                        kz[-1], mixl[-1], gravity, mw_atmos, gas_mw[i], q_below,
                        supsat, fsed, b, eps, z_bot, z_base, z_alpha, z_min, param,
                        sig,mh, rmin, nrad, d_molecule,eps_k,c_p_factor, #all scalaers
                        og_vfall, z_cld
                    )

        z_cld=None
        for iz in range(nz-1,-1,-1): #goes from BOA to TOA

            qc[iz,i], qt[iz,i], rg[iz,i], reff[iz,i],ndz[iz,i],q_below, z_cld, fsed_layer[iz,i]  = layer( igas, rho_p[i], 
                #t,p layers, then t.p levels below and above
                t_mid[iz], p_mid[iz], t_top[iz], t_top[iz+1], p_top[iz], p_top[iz+1],
                kz[iz], mixl[iz], gravity, mw_atmos, gas_mw[i], q_below,  
                supsat, fsed, b, eps, z_top[iz], z_top[iz+1], z_alpha, z_min, param,
                sig,mh, rmin, nrad, d_molecule,eps_k,c_p_factor, #all scalars
                og_vfall, z_cld
            )

            qc_path[i] = (qc_path[i] + qc[iz,i]*
                            ( p_top[iz+1] - p_top[iz] ) / gravity)
        z_cld_out[i] = z_cld

    return qc, qt, rg, reff, ndz, qc_path,mixl, z_cld_out

def layer(gas_name,rho_p, t_layer, p_layer, t_top, t_bot, p_top, p_bot,
    kz, mixl, gravity, mw_atmos, gas_mw, q_below,
    supsat, fsed, b, eps, z_top, z_bot, z_alpha, z_min, param,
    sig,mh, rmin, nrad, d_molecule,eps_k,c_p_factor,
    og_vfall, z_cld):
    """
    Calculate layer condensate properties by iterating on optical depth
    in one model layer (convering on optical depth over sublayers)

    gas_name : str 
        Name of condenstante 
    rho_p : float 
        density of condensed vapor (g/cm^3)
    t_layer : float 
        Temperature of layer mid-pt (K)
    p_layer : float 
        Pressure of layer mid-pt (dyne/cm^2)
    t_top : float 
        Temperature at top of layer (K)
    t_bot : float 
        Temperature at botton of layer (K)
    p_top : float 
        Pressure at top of layer (dyne/cm2)
    p_bot : float 
        Pressure at botton of layer 
    kz : float 
        eddy diffusion coefficient (cm^2/s)
    mixl : float 
        Mixing length (cm)
    gravity : float 
        Gravity of planet cgs 
    mw_atmos : float 
        Molecular weight of the atmosphere 
    gas_mw : float 
        Gas molecular weight 
    q_below : float 
        total mixing ratio (vapor+condensate) below layer (g/g)
    supsat : float 
        Super saturation factor
    fsed : float
        Sedimentation efficiency coefficient (unitless) 
    b : float
        Denominator of exponential in sedimentation efficiency  (if param is 'exp')
    eps: float
        Minimum value of fsed function (if param=exp)
    z_top : float
        Altitude at top of layer
    z_bot : float
        Altitude at bottom of layer
    z_alpha : float
        Altitude at which fsed=alpha for variable fsed calculation
    param : str
        fsed parameterisation
        'const' (constant), 'exp' (exponential density derivation)
    sig : float 
        Width of the log normal particle distribution 
    mh : float 
        Metallicity NON log soar (1=1xSolar)
    rmin : float 
        Minium radius on grid (cm)
    nrad : int 
        Number of radii on Mie grid
    d_molecule : float 
        diameter of atmospheric molecule (cm) (Rosner, 2000)
        (3.711e-8 for air, 3.798e-8 for N2, 2.827e-8 for H2)
        Set in Atmosphere constants 
    eps_k : float 
        Depth of the Lennard-Jones potential well for the atmosphere 
        Used in the viscocity calculation (units are K) (Rosner, 2000)
    c_p_factor : float 
        specific heat of atmosphere (erg/K/g) . Usually 7/2 for ideal gas
        diatomic molecules (e.g. H2, N2). Technically does slowly rise with 
        increasing temperature
    og_vfall : bool 
        Use original or new vfall calculation

    Returns
    -------
    qc_layer : ndarray 
        condenstate mixing ratio (g/g)
    qt_layer : ndarray 
        gas + condensate mixing ratio (g/g)
    rg_layer : ndarray
        geometric mean radius of condensate  cm 
    reff_layer : ndarray
        droplet effective radius (second moment of size distrib, cm)
    ndz_layer : ndarray 
        number column density of condensate (cm^-3)
    q_below : ndarray 
        total mixing ratio (vapor+condensate) below layer (g/g)
    """
    #   universal gas constant (erg/mol/K)
    nsub_max = 128
    R_GAS = 8.3143e7
    AVOGADRO = 6.02e23
    K_BOLTZ = R_GAS / AVOGADRO
    PI = np.pi 
    #   Number of levels of grid refinement used 
    nsub = 1

    #   specific gas constant for atmosphere (erg/K/g)
    r_atmos = R_GAS / mw_atmos

    #specific gas constant for cloud (erg/K/g)
    r_cloud = R_GAS/ gas_mw

    #   specific heat of atmosphere (erg/K/g)
    c_p = c_p_factor * r_atmos

    #   pressure thickness of layer
    dp_layer = p_bot - p_top
    dlnp = np.log( p_bot/p_top )

    #   temperature gradient 
    dtdlnp = ( t_top - t_bot ) / dlnp
    lapse_ratio = ( t_bot - t_top ) / dlnp / ( t_layer / c_p_factor )

    #   atmospheric density (g/cm^3)
    rho_atmos = p_layer / ( r_atmos * t_layer )

    #   atmospheric scale height (cm)
    scale_h = r_atmos * t_layer / gravity    

    #   convective velocity scale (cm/s) from mixing length theory
    w_convect = kz / mixl 

    #   atmospheric number density (molecules/cm^3)
    n_atmos = p_layer / ( K_BOLTZ*t_layer )

    #   atmospheric mean free path (cm)
    mfp = 1. / ( np.sqrt(2.)*n_atmos*PI*d_molecule**2 )

    # atmospheric viscosity (dyne s/cm^2)
    # EQN B2 in A & M 2001, originally from Rosner+2000
    # Rosner, D. E. 2000, Transport Processes in Chemically Reacting Flow Systems (Dover: Mineola)
    visc = (5./16.*np.sqrt( PI*K_BOLTZ*t_layer*(mw_atmos/AVOGADRO)) /
        ( PI*d_molecule**2 ) /
        ( 1.22 * ( t_layer / eps_k )**(-0.16) ))


    #   --------------------------------------------------------------------
    #   Top of convergence loop    
    converge = False
    while not converge: 

        #   Zero cumulative values
        qc_layer = 0.
        qt_layer = 0.
        ndz_layer = 0.
        opd_layer = 0.        

        #   total mixing ratio and pressure at bottom of sub-layer

        qt_bot_sub = q_below
        p_bot_sub = p_bot
        z_bot_sub = z_bot

        #SUBALYER 
        dp_sub = dp_layer / nsub

        for isub in range(nsub): 
            qt_below = qt_bot_sub
            p_top_sub = p_bot_sub - dp_sub
            dz_sub = scale_h * np.log( p_bot_sub/p_top_sub ) # width of layer
            p_sub = 0.5*( p_bot_sub + p_top_sub )
            #################### CHECK #####################
            z_top_sub = z_bot_sub + dz_sub
            z_sub = z_bot_sub + scale_h * np.log( p_bot_sub/p_sub ) # midpoint of layer 
            ################################################
            t_sub = t_bot + np.log( p_bot/p_sub )*dtdlnp
            qt_top, qc_sub, qt_sub, rg_sub, reff_sub,ndz_sub, z_cld, fsed_layer = calc_qc(
                    gas_name, supsat, t_sub, p_sub,r_atmos, r_cloud,
                        qt_below, mixl, dz_sub, gravity,mw_atmos,mfp,visc,
                        rho_p,w_convect, fsed, b, eps, param, z_bot_sub, z_sub, z_alpha, z_min,
                        sig,mh, rmin, nrad, og_vfall,z_cld)


            #   vertical sums
            qc_layer = qc_layer + qc_sub*dp_sub/gravity
            qt_layer = qt_layer + qt_sub*dp_sub/gravity
            ndz_layer = ndz_layer + ndz_sub

            if reff_sub > 0.:
                opd_layer = (opd_layer + 
                                    1.5*qc_sub*dp_sub/gravity/(rho_p*reff_sub))
    
            #   Increment values at bottom of sub-layer

            qt_bot_sub = qt_top
            p_bot_sub = p_top_sub
            z_bot_sub = z_top_sub

        #    Check convergence on optical depth
        if nsub_max == 1 :
            converge = True
        elif  nsub == 1 : 
            opd_test = opd_layer
        elif (opd_layer == 0.) or (nsub >= nsub_max): 
            converge = True
        elif ( abs( 1. - opd_test/opd_layer ) <= 1e-2 ) : 
            converge = True
        else: 
            opd_test = opd_layer
        
        nsub = nsub * 2
    #   Update properties at bottom of next layer

    q_below = qt_top

    #Get layer averages

    if opd_layer > 0. : 
        reff_layer = 1.5*qc_layer / (rho_p*opd_layer)
        lnsig2 = 0.5*np.log( sig )**2
        rg_layer = reff_layer*np.exp( -5*lnsig2 )
    else : 
        reff_layer = 0.
        rg_layer = 0.

    qc_layer = qc_layer*gravity / dp_layer
    qt_layer = qt_layer*gravity / dp_layer

    return qc_layer, qt_layer, rg_layer, reff_layer, ndz_layer,q_below, z_cld, fsed_layer

def calc_qc(gas_name, supsat, t_layer, p_layer
    ,r_atmos, r_cloud, q_below, mixl, dz_layer, gravity,mw_atmos
    ,mfp,visc,rho_p,w_convect, fsed, b, eps, param, z_bot, z_layer, z_alpha, z_min,
    sig, mh, rmin, nrad, og_vfall=True, z_cld=None):
    """
    Calculate condensate optical depth and effective radius for a layer,
    assuming geometric scatterers. 

    gas_name : str 
        Name of condensate 
    supsat : float 
        Super saturation factor 
    t_layer : float 
        Temperature of layer mid-pt (K)
    p_layer : float 
        Pressure of layer mid-pt (dyne/cm^2)
    r_atmos : float 
        specific gas constant for atmosphere (erg/K/g)
    r_cloud : float 
        specific gas constant for cloud species (erg/K/g)     
    q_below : float 
        total mixing ratio (vapor+condensate) below layer (g/g)
    mxl : float 
        convective mixing length scale (cm): no less than 1/10 scale height
    dz_layer : float 
        Thickness of layer cm 
    gravity : float 
        Gravity of planet cgs 
    mw_atmos : float 
        Molecular weight of the atmosphere 
    mfp : float 
        atmospheric mean free path (cm)
    visc : float 
        atmospheric viscosity (dyne s/cm^2)
    rho_p : float 
        density of condensed vapor (g/cm^3)
    w_convect : float    
        convective velocity scale (cm/s)
    fsed : float
        Sedimentation efficiency coefficient (unitless) 
    b : float
        Denominator of exponential in sedimentation efficiency  (if param is 'exp')
    eps: float
        Minimum value of fsed function (if param=exp)
    z_bot : float
        Altitude at bottom of layer
    z_layer : float 
        Altitude of midpoint of layer (cm)
    z_alpha : float
        Altitude at which fsed=alpha for variable fsed calculation
    param : str
        fsed parameterisation
        'const' (constant), 'exp' (exponential density derivation)
    sig : float 
        Width of the log normal particle distrubtion 
    mh : float 
        Metallicity NON log solar (1 = 1x solar)
    rmin : float 
        Minium radius on grid (cm)
    nrad : int 
        Number of radii on Mie grid

    Returns
    -------
    qt_top : float 
        gas + condensate mixing ratio at top of layer(g/g)
    qc_layer : float 
        condenstate mixing ratio (g/g)
    qt_layer : float 
        gas + condensate mixing ratio (g/g)
    rg_layer : float
        geometric mean radius of condensate  cm 
    reff_layer : float
        droplet effective radius (second moment of size distrib, cm)
    ndz_layer : float 
        number column density of condensate (cm^-3)
    """

    get_pvap = getattr(pvaps, gas_name)
    if gas_name == 'Mg2SiO4':
        pvap = get_pvap(t_layer, p_layer,mh=mh)
    else:
        pvap = get_pvap(t_layer,mh=mh)

    fs = supsat + 1 

    #   atmospheric density (g/cm^3)
    rho_atmos = p_layer / ( r_atmos * t_layer )    

    #   mass mixing ratio of saturated vapor (g/g)
    qvs = fs*pvap / ( (r_cloud) * t_layer ) / rho_atmos

    #   --------------------------------------------------------------------
    #   Layer is cloud free

    if( q_below < qvs ):

        qt_layer = q_below
        qt_top   = q_below
        qc_layer = 0.
        rg_layer = 0.
        reff_layer = 0.
        ndz_layer = 0.
        z_cld = z_cld
        fsed_mid = 0

    else:

        #   --------------------------------------------------------------------
        #   Cloudy layer: first calculate qt and qc at top of layer,
        #   then calculate layer averages
        if z_cld is None:
            z_cld = z_bot
        else:
            z_cld = z_cld

        #   range of mixing ratios to search (g/g)
        qhi = q_below
        qlo = qhi / 1e3

        #   load parameters into advdiff common block

        ad_qbelow = q_below
        ad_qvs = qvs
        ad_mixl = mixl
        ad_dz = dz_layer
        ad_rainf = fsed

        #   Find total vapor mixing ratio at top of layer
        #find_root = True
        #while find_root:
        #    try:
        #        qt_top = optimize.root_scalar(advdiff, bracket=[qlo, qhi], method='brentq',
        #                    args=(ad_qbelow,ad_qvs, ad_mixl,ad_dz ,ad_rainf,
        #                        z_bot, b, eps, param)
        #                        )#, xtol = 1e-20)
        #        find_root = False
        #    except ValueError:
        #        qlo = qlo/10
        #
        #qt_top = qt_top.root
        if param is "const":
            qt_top = qvs + (q_below - qvs) * np.exp(-fsed * dz_layer / mixl)
        elif param is "exp":
            fs = fsed / np.exp(z_alpha / b)
            qt_top = qvs + (q_below - qvs) * np.exp( - b * fs / mixl * np.exp(z_bot/b) 
                            * (np.exp(dz_layer/b) -1) + eps*dz_layer/mixl)

        #   Use trapezoid rule (for now) to calculate layer averages
        #   -- should integrate exponential
        qt_layer = 0.5*( q_below + qt_top )


        #   Find total condensate mixing ratio
        qc_layer = np.max( [0., qt_layer - qvs] )

        #   --------------------------------------------------------------------
        #   Find <rw> corresponding to <w_convect> using function vfall()

        #   range of particle radii to search (cm)
        rlo = 1.e-10
        rhi = 10.
        
        #   precision of vfall solution (cm/s)
        find_root = True
        while find_root:
            try:
                if og_vfall:
                    rw_temp = optimize.root_scalar(vfall_find_root, bracket=[rlo, rhi], method='brentq', 
                            args=(gravity,mw_atmos,mfp,visc,t_layer,p_layer, rho_p,w_convect))
                else:
                    rw_temp = solve_force_balance("rw", w_convect, gravity, mw_atmos, mfp,
                                                    visc, t_layer, p_layer, rho_p, rlo, rhi)
                find_root = False
            except ValueError:
                rlo = rlo/10
                rhi = rhi*10

        #fall velocity particle radius 
        if og_vfall: rw_layer = rw_temp.root
        else: rw_layer = rw_temp
        
        #   geometric std dev of lognormal size distribution
        lnsig2 = 0.5*np.log( sig )**2
        #   sigma floor for the purpose of alpha calculation
        sig_alpha = np.max( [1.1, sig] )    

        #   find alpha for power law fit vf = w(r/rw)^alpha
        def pow_law(r, alpha):
            return np.log(w_convect) + alpha * np.log (r / rw_layer) 

        r_, rup, dr = get_r_grid(r_min = rmin, n_radii = nrad)
        vfall_temp = []
        for j in range(len(r_)):
            if og_vfall:
                vfall_temp.append(vfall(r_[j], gravity, mw_atmos, mfp, visc, t_layer, p_layer, rho_p))
            else:
                vlo = 1e0; vhi = 1e6
                find_root = True
                while find_root:
                    try:
                        vfall_temp.append(solve_force_balance("vfall", r_[j], gravity, mw_atmos, 
                            mfp, visc, t_layer, p_layer, rho_p, vlo, vhi))
                        find_root = False
                    except ValueError:
                        vlo = vlo/10
                        vhi = vhi*10

        pars, cov = optimize.curve_fit(f=pow_law, xdata=r_, ydata=np.log(vfall_temp), p0=[0], 
                            bounds=(-np.inf, np.inf))
        alpha = pars[0]


        #   fsed at middle of layer 
        if param is 'exp':
            fsed_mid = fs * np.exp(z_layer / b) + eps
        else: # 'const'
            fsed_mid = fsed

        #     EQN. 13 A&M 
        #   geometric mean radius of lognormal size distribution
        rg_layer = (fsed_mid**(1./alpha) *
                    rw_layer * np.exp( -(alpha+6)*lnsig2 ))

        #   droplet effective radius (cm)
        reff_layer = rg_layer*np.exp( 5*lnsig2 )

        #      EQN. 14 A&M
        #   column droplet number concentration (cm^-2)
        ndz_layer = (3*rho_atmos*qc_layer*dz_layer /
                    ( 4*np.pi*rho_p*rg_layer**3 ) * np.exp( -9*lnsig2 ))

    return qt_top, qc_layer,qt_layer, rg_layer,reff_layer,ndz_layer, z_cld, fsed_mid 

class Atmosphere():
    def __init__(self,condensibles, fsed=0.5, b=1, eps=1e-2, mh=1, mmw=2.2, sig=2.0,
                    param='const', verbose=True, supsat=0, gas_mmr=None):
        """
        Parameters
        ----------
        condensibles : list of str
            list of gases for which to consider as cloud species 
        fsed : float 
            Sedimentation efficiency coefficient. Jupiter ~3-6. Hot Jupiters ~ 0.1-1.
        b : float
            Denominator of exponential in sedimentation efficiency  (if param is 'exp')
        eps: float
            Minimum value of fsed function (if param=exp)
        mh : float 
            metalicity 
        mmw : float 
            MMW of the atmosphere 
        sig : float 
            Width of the log normal distribution for the particle sizes 
        param : str
            fsed parameterisation
            'const' (constant), 'exp' (exponential density derivation)
        verbose : bool 
            Prints out warning statements throughout
    
        """
        self.mh = mh
        self.mmw = mmw
        self.condensibles = condensibles
        self.fsed = fsed
        self.b = b
        self.sig = sig
        self.param = param
        self.eps = eps
        self.verbose = verbose 
        #grab constants
        self.constants()
        self.supsat = supsat
        self.gas_mmr = gas_mmr

    def constants(self):
        #   Depth of the Lennard-Jones potential well for the atmosphere 
        # Used in the viscocity calculation (units are K) (Rosner, 2000)
        #   (78.6 for air, 71.4 for N2, 59.7 for H2)
        self.eps_k = 59.7
        #   diameter of atmospheric molecule (cm) (Rosner, 2000)
        #   (3.711e-8 for air, 3.798e-8 for N2, 2.827e-8 for H2)
        self.d_molecule = 2.827e-8

        #specific heat factor of the atmosphere 
        #7/2 comes from ideal gas assumption of permanent diatomic gas 
        #e.g. h2, o2, n2, air, no, co
        #this technically does increase slowly toward higher temperatures (>~700K)
        self.c_p_factor = 7./2.

        self.R_GAS = 8.3143e7
        self.AVOGADRO = 6.02e23
        self.K_BOLTZ = self.R_GAS / self.AVOGADRO

    def ptk(self, df = None, filename=None,
        kz_min=1e5, constant_kz=None, latent_heat=False, convective_overshoot=None,
        Teff=None, alpha_pressure=None, **pd_kwargs):
        """
        Read in file or define dataframe. 
    
        Parameters
        ----------
        df : dataframe or dict
            Dataframe with "pressure"(bars),"temperature"(K). MUST have at least two 
            columns with names "pressure" and "temperature". 
            Optional columns include the eddy diffusion "kz" in cm^2/s CGS units, and 
            the convective heat flux 'chf' also in cgs (e.g. sigma_csg T^4)
        filename : str 
            Filename read in. Will be read in with pd.read_csv and should 
            result in two named headers "pressure"(bars),"temperature"(K). 
            Optional columns include the eddy diffusion "kz" in cm^2/s CGS units, and 
            the convective heat flux 'chf' also in cgs (e.g. sigma_csg T^4)
            Use pd_kwargs to ensure file is read in properly.
        kz_min : float, optional
            Minimum Kz value. This will reset everything below kz_min to kz_min. 
            Default = 1e5 cm2/s
        constant_kz : float, optional
            Constant value for kz, if kz is supplied in df or filename, 
            it will inheret that value and not use this constant_value
            Default = None 
        latent_heat : bool 
            optional, Default = False. The latent heat factors into the mixing length. 
            When False, the mixing length goes as the scale height 
            When True, the mixing length is scaled by the latent heat 
        convective_overshoot : float 
            Optional, Default is None. But the default value used in 
            Ackerman & Marley 2001 is 1./3. If you are unsure of what to pick, start 
            there. This is only used when the        
            This is ONLY used when a chf (convective heat flux) is supplied 
        Teff : float, optional
            Effective temperature. If None (default), Teff set to temperature at 1 bar
        alpha_pressure : float
            Pressure at which we want fsed=alpha for variable fsed calculation
        pd_kwargs : kwargs
            Pandas key words for file read in. 
            If reading old style eddysed files, you would need: 
            skiprows=3, delim_whitespace=True, header=None, names=["ind","pressure","temperature","kz"]
        """
        #first read in dataframe, dict or file and sort by pressure
        if not isinstance(df, type(None)):
            if isinstance(df, dict): df = pd.DataFrame(df)
            df = df.sort_values('pressure')
        elif not isinstance(filename, type(None)):
            df = pd.read_csv(filename, **pd_kwargs)
            df = df.sort_values('pressure')

        #convert bars to dyne/cm^2 
        self.p_level = np.array(df['pressure'])*1e6
        self.t_level = np.array(df['temperature'])      
        if alpha_pressure is None:
            self.alpha_pressure=min(df['pressure'])
        else:
            self.alpha_pressure=alpha_pressure
        self.get_atmo_parameters()
        self.get_kz_mixl(df, constant_kz, latent_heat, convective_overshoot, kz_min)

        # Teff
        if isinstance(Teff, type(None)):
            onebar = (np.abs(self.p_level/1e6 - 1.)).argmin()
            self.Teff = self.t_level[onebar]
        else:
            self.Teff = Teff



    def get_atmo_parameters(self):
        """Defines all the atmospheric parameters needed for the calculation

        Note: Some of this is repeated in the layer() function. 
        This is done on purpose because layer is used to get a "virtual"
        layer off the user input's grid. These parameters are also 
        needed, though, to get the initial mixing parameters from the user. 
        Therefore, we put them in both places. 
        """
        #   specific gas constant for atmosphere (erg/K/g)
        self.r_atmos = self.R_GAS / self.mmw
        
        # pressure thickess 
        dlnp = np.log( self.p_level[1:] / self.p_level[0:-1] ) #USED IN LAYER
        self.p_layer = 0.5*( self.p_level[1:] + self.p_level[0:-1]) #USED IN LAYER
        
        #   temperature gradient - we use this for the sub layering
        self.dtdlnp = ( self.t_level[0:-1] - self.t_level[1:] ) / dlnp #USED IN LAYER
        
        # get temperatures at layers
        self.t_layer = self.t_level[1:] + np.log( self.p_level[1:]/self.p_layer )*self.dtdlnp #USED IN LAYER

        #lapse ratio used for kz calculation if user asks for 
        # us ot calculate it based on convective heat flux
        self.lapse_ratio = ( self.t_level[1:] - self.t_level[0:-1] 
                            ) / dlnp / ( self.t_layer/self.c_p_factor )

        #   atmospheric density (g/cm^3)
        self.rho_atmos = self.p_layer / (self.r_atmos * self.t_layer)

        #scale height (todo=make this gravity dependent)
        self.scale_h = self.r_atmos * self.t_layer / self.g

        #specific heat of atmosphere
        self.c_p = self.c_p_factor * self.r_atmos

        #get altitudes 
        self.dz_pmid = self.scale_h * np.log( self.p_level[1:]/self.p_layer )

        self.dz_layer = self.scale_h * dlnp

        self.z_top = np.concatenate(([0],np.cumsum(self.dz_layer[::-1])))[::-1]

        self.z = self.z_top[1:]+self.dz_pmid

        # altitude to set fsed = alpha
        p_alpha = find_nearest_1d(self.p_layer/1e6, self.alpha_pressure)
        z_temp = np.cumsum(self.dz_layer[::-1])[::-1]
        self.z_alpha = z_temp[p_alpha]

    def get_kz_mixl(self, df, constant_kz, latent_heat, convective_overshoot,
     kz_min):
        """
        Computes kz profile and mixing length given user input. In brief the options are: 

        1) Input Kz
        2) Input constant kz
        3) Input convective heat flux (supply chf in df)
        3a) Input convective heat flux, correct for latent heat (supply chf in df and set latent_heat=True)
        and/or 3b) Input convective heat flux, correct for convective overshoot (supply chf, convective_overshoot=1/3)
        4) Set kz_min to prevent kz from going too low (any of the above and set kz_min~1e5)

        Parameters
        ----------
        df : dataframe or dict
            Dataframe from input with "pressure"(bars),"temperature"(K). MUST have at least two 
            columns with names "pressure" and "temperature". 
            Optional columns include the eddy diffusion "kz" in cm^2/s CGS units, and 
            the convective heat flux 'chf' also in cgs (e.g. sigma_csg T^4)
        constant_kz : float
            Constant value for kz, if kz is supplied in df or filename, 
            it will inheret that value and not use this constant_value
            Default = None 
        latent_heat : bool 
            optional, Default = False. The latent heat factors into the mixing length. 
            When False, the mixing length goes as the scale height 
            When True, the mixing length is scaled by the latent heat 
        convective_overshoot : float 
            Optional, Default is None. But the default value used in 
            Ackerman & Marley 2001 is 1./3. If you are unsure of what to pick, start 
            there. This is only used when the        
            This is ONLY used when a chf (convective heat flux) is supplied 
        kz_min : float
            Minimum Kz value. This will reset everything below kz_min to kz_min. 
            Default = 1e5 cm2/s
        """

        #MIXING LENGTH ASSUMPTIONS 
        if latent_heat:
            #   convective mixing length scale (cm): no less than 1/10 scale height
            self.mixl = np.array([np.max( [0.1, ilr] ) for ilr in self.lapse_ratio]) * self.scale_h
        else:
            #convective mixing length is the scale height
            self.mixl = 1 * self.scale_h


        #KZ OPTIONS 

        #   option 1) the user has supplied it in their file or dictionary
        if 'kz' in df.keys(): 
            if df.loc[df['kz']<kz_min].shape[0] > 0:
                df.loc[df['kz']<kz_min] = kz_min
                if self.verbose: print('Overwriting some Kz values to minimum value set by kz_min \n \
                    You can always turn off these warnings by setting verbose=False') 
            kz_level = np.array(df['kz'])
            self.kz = 0.5*(kz_level[1:] + kz_level[0:-1])
            self.chf = None

        #   option 2) the user wants a constant value
        elif not isinstance(constant_kz , type(None)):
            self.kz = np.zeros(df.shape[0]-1) + constant_kz
            self.chf = None

        #   option 3) the user wants to compute kz based on a convective heat flux 
        elif 'chf' in df.keys():
            self.chf =  np.array(df['chf'])


            #CONVECTIVE OVERSHOOT ON OR OFF
            #     sets the minimum allowed heat flux in a layer by assuming some overshoot
            #     the default value of 1/3 is arbitrary, allowing convective flux to fall faster than
            #     pressure scale height
            if not isinstance(convective_overshoot, type(None)):
                used = False
                nz = len(self.p_layer)
                for iz in range(nz-1,-1,-1):
                    ratio_min = (convective_overshoot)*self.p_level[iz]/self.p_level[iz+1] 
                    if self.chf[iz] < ratio_min*self.chf[iz+1]:
                        self.chf[iz] = self.chf[iz+1]*ratio_min
                        used=True
                if self.verbose: print('Convective overshoot was turned on. The convective heat flux \n \
                    has been adjusted such that it is not allowed to decrease more than {0} \n \
                    the pressure. This number is set with the convective_overshoot parameter. \n \
                    It can be disabled with convective_overshoot=None. To turn \n \
                    off these messages set verbose=False in Atmosphere'.format(convective_overshoot)) 

            #   vertical eddy diffusion coefficient (cm^2/s)
            #   from Gierasch and Conrath (1985)
            gc_kzz = ((1./3.) * self.scale_h * (self.mixl/self.scale_h)**(4./3.) * 
                    ( ( self.r_atmos*self.chf[1:] ) / ( self.rho_atmos*self.c_p  ) )**(1./3.)) 
            
            self.kz =  [np.max([i, kz_min]) for i in gc_kzz ]
        else:
            raise Exception("Users can define kz by: \n \
            1) Adding 'kz' as a column or key to your dataframe dict, or file \n \
            2) Defining constant-w-altitude kz through the constant_kz input \n  \
            3) Adding 'chf', the conective heat flux as a column to your \
            dataframe, dict or file.")

    def gravity(self, gravity=None, gravity_unit=None, radius=None, 
        radius_unit=None, mass = None, mass_unit=None):
        """
        Get gravity based on mass and radius, or gravity inputs 

        Parameters
        ----------
        gravity : float 
            (Optional) Gravity of planet 
        gravity_unit : astropy.unit
            (Optional) Unit of Gravity
        radius : float 
            (Optional) radius of planet MUST be specified for thermal emission!
        radius_unit : astropy.unit
            (Optional) Unit of radius
        mass : float 
            (Optional) mass of planet 
        mass_unit : astropy.unit
            (Optional) Unit of mass   
        """
        if (mass is not None) and (radius is not None):
            m = (mass*mass_unit).to(u.g)
            r = (radius*radius_unit).to(u.cm)
            g = (c.G.cgs * m /  (r**2)).value
            self.g = g
            self.gravity_unit = 'cm/(s**2)'
        elif gravity is not None:
            g = (gravity*gravity_unit).to('cm/(s**2)')
            g = g.value
            self.g = g
            self.gravity_unit = 'cm/(s**2)'
        else: 
            raise Exception('Need to specify gravity or radius and mass + additional units')


    def kz(self,df = None, constant_kz=None, chf = None, kz_min = 1e5, latent_heat=False): 
        """
        Define Kz in CGS. Should be on same grid as pressure. This overwrites whatever was 
        defined in get_pt ! Users can define kz by: 
            1) Defining a DataFrame with keys 'pressure' (in bars), and 'kz'
            2) Defining constant kz 
            3) Supplying a convective heat flux and prescription for latent_heat

        Parameters
        ----------
        df : pandas.DataFrame, dict
            Dataframe or dictionary with 'kz' as one of the fields. 
        constant_kz : float 
            Constant value for kz in units of cm^2/s
        chf : ndarray 
            Convective heat flux in cgs units (e.g. sigma T^4). This will be used to compute 
            the kzz using the methodology of Gierasch and Conrath (1985)
        latent_heat : bool 
            optional, Default = False. The latent heat factors into the mixing length. 
            When False, the mixing length goes as the scale height 
            When True, the mixing length is scaled by the latent heat 
            This is ONLY used when a chf (convective heat flux) is supplied 
        """
        return "Depricating this function. Please use ptk instead. It has identical functionality."
        if not isinstance(df, type(None)):
            #will not need any convective heat flux 
            self.chf = None
            #reset to minimun value if specified by the user
            if df.loc[df['kz']<kz_min].shape[0] > 0:
                df.loc[df['kz']<kz_min] = kz_min
                print('Overwriting some Kz values to minimum value set by kz_min') 
            self.kz = np.array(df['kz'])
            #make sure pressure and kz are the same size 
            if len(self.kz) != len(self.pressure) : 
                raise Exception('Kzz and pressure are not the same length')

        elif not isinstance(constant_kz, type(None)):
            #will not need any convective heat flux
            self.chf = None
            self.kz = constant_kz
            if self.kz<kz_min:
                self.kz = kz_min
                print('Overwriting kz constant value to minimum value set by kz_min')

        elif not isinstance(chf, type(None)):
            def g_c_85(scale_h,r_atmos, chf, rho_atmos, c_p, lapse_ratio):
                #   convective mixing length scale (cm): no less than 1/10 scale height
                if latent_heat:
                    mixl = np.max( 0.1, lapse_ratio ) * scale_h
                else:
                    mixl = scale_h
                #   vertical eddy diffusion coefficient (cm^2/s)
                #   from Gierasch and Conrath (1985)
                gc_kzz = ((1./3.) * scale_h * (mixl/scale_h)**(4./3.) * 
                        ( ( r_atmos*chf ) / ( rho_atmos*c_p ) )**(1./3.)) 
                return np.max(gc_kzz, kz_min), mixl

            self.kz = g_c_85
            self.chf = chf

    def compute(self,directory = None, as_dict = True): 
        """
        Parameters
        ----------
        atmo : class 
            `Atmosphere` class 
        directory : str, optional 
            Directory string that describes where refrind files are 
        as_dict : bool 
            Default = True, option to view full output as dictionary

        Returns 
        -------
        dict 
            When as_dict=True. Dictionary output that contains full output. See tutorials for explanation of all output.        
        opd, w0, g0
            Extinction per layer, single scattering abledo, asymmetry parameter, 
            All are ndarrays that are nlayer by nwave
        """
        run = compute(self, directory = directory, as_dict = as_dict)
        return run

def calc_mie_db(gas_name, dir_refrind, dir_out, rmin = 1e-8, nradii = 60):
    """
    Function that calculations new Mie database using PyMieScatt. 

    Parameters
    ----------
    gas_name : list, str
        List of names of gasses. Or a single gas name. 
        See pyeddy.available() to see which ones are currently available. 
    dir_refrind : str 
        Directory where you store optical refractive index files that will be created. 
    dir_out: str 
        Directory where you want to store Mie parameter files. Will be stored as gas_name.Mieff. 
        BEWARE FILE OVERWRITES. 
    rmin : float , optional
        (Default=1e-5) Units of cm. The minimum radius to compute Mie parameters for. 
        Usually 0.1 microns is small enough. However, if you notice your mean particle radius 
        is on the low end, you may compute your grid to even lower particle sizes. 
    nradii : int, optional
        (Default=40) number of radii points to compute grid on. 40 grid points for exoplanets/BDs
        is generally sufficient. 

    Returns 
    -------
    Q extinction, Q scattering,  asymmetry * Q scattering, radius grid (cm), wavelength grid (um)

    The Q "efficiency factors" are = cross section / geometric cross section of particle
    """
    if isinstance(gas_name,str):
        gas_name = [gas_name]
    ngas = len(gas_name)

    for i in range(len(gas_name)): 
        print('Computing ' + gas_name[i])
        #Setup up a particle size grid on first run and calculate single-particle scattering
        
        #files will be saved in `directory`
        # obtaining refractive index data for each gas
        wave_in,nn,kk = get_refrind(gas_name[i],dir_refrind)
        nwave = len(wave_in)

        if i==0:
            #all these files need to be on the same grid
            radius, rup, dr = get_r_grid(r_min = rmin, n_radii = nradii)

            qext_all=np.zeros(shape=(nwave,nradii,ngas))
            qscat_all = np.zeros(shape=(nwave,nradii,ngas))
            cos_qscat_all=np.zeros(shape=(nwave,nradii,ngas))

        #get extinction, scattering, and asymmetry
        #all of these are  [nwave by nradii]
        qext_gas, qscat_gas, cos_qscat_gas = calc_new_mieff(wave_in, nn,kk, radius, rup, fort_calc_mie = False)

        #add to master matrix that contains the per gas Mie stuff
        qext_all[:,:,i], qscat_all[:,:,i], cos_qscat_all[:,:,i] = qext_gas, qscat_gas, cos_qscat_gas 

        #prepare format for old ass style
        wave = [nwave] + sum([[r]+list(wave_in) for r in radius],[])
        qscat = [nradii]  + sum([[np.nan]+list(iscat) for iscat in qscat_gas.T],[])
        qext = [np.nan]  + sum([[np.nan]+list(iext) for iext in qext_gas.T],[])
        cos_qscat = [np.nan]  + sum([[np.nan]+list(icos) for icos in cos_qscat_gas.T],[])

        pd.DataFrame({'wave':wave,'qscat':qscat,'qext':qext,'cos_qscat':cos_qscat}).to_csv(os.path.join(dir_out,gas_name[i]+".mieff"),
                                                                                   sep=' ',
                                                                                  index=False,header=None)
    return qext_all, qscat_all, cos_qscat_all, radius,wave_in

def get_mie(gas, directory):
    """
    Get Mie parameters from old ass formatted files
    """
    df = pd.read_csv(os.path.join(directory,gas+".mieff"),names=['wave','qscat','qext','cos_qscat'], delim_whitespace=True)

    nwave = int( df.iloc[0,0])
    nradii = int(df.iloc[0,1])

    #get the radii (all the rows where there the last three rows are nans)
    radii = df.loc[np.isnan(df['qscat'])]['wave'].values

    df = df.dropna()

    assert len(radii) == nradii , "Number of radii specified in header is not the same as number of radii."
    assert nwave*nradii == df.shape[0] , "Number of wavelength specified in header is not the same as number of waves in file"

    wave = df['wave'].values.reshape((nradii,nwave)).T
    qscat = df['qscat'].values.reshape((nradii,nwave)).T
    qext = df['qext'].values.reshape((nradii,nwave)).T
    cos_qscat = df['cos_qscat'].values.reshape((nradii,nwave)).T

    return qext,qscat, cos_qscat, nwave, radii,wave

def get_refrind(igas,directory): 
    """
    Reads reference files with wavelength, and refractory indecies. 
    This function relies on input files being structured as a 4 column file with 
    columns: index, wavelength (micron), nn, kk 

    Parameters
    ----------
    igas : str 
        Gas name 
    directory : str 
        Directory were reference files are located. 
    """
    filename = os.path.join(directory ,igas+".refrind")
    #put skiprows=1 in loadtxt to skip first line
    idummy, wave_in, nn, kk = np.loadtxt(open(filename,'rt').readlines(), unpack=True, usecols=[0,1,2,3])#[:-1]

    return wave_in,nn,kk

def get_r_grid_w_max(r_min=1e-8, r_max=5.4239131e-2, n_radii=60):
    """
    Get spacing of radii to run Mie code

    r_min : float 
            Minimum radius to compute (cm)
    r_max : float 
            Maximum radius to compute (cm)
    n_radii : int
            Number of radii to compute 
    """

    radius = np.logspace(np.log10(r_min),np.log10(r_max),n_radii)
    rat = radius[1]/radius[0]
    rup = 2*rat / (rat+1) * radius
    dr = np.zeros(rup.shape)
    dr[1:] = rup[1:]-rup[:-1]
    dr[0] = dr[1]**2/dr[2]

    return radius, rup, dr

def get_r_grid(r_min=1e-8, n_radii=60):
    """
    Warning
    -------
    Original code from A&M code. 
    Discontinued function. See 'get_r_grid'.

    Get spacing of radii to run Mie code

    r_min : float 
        Minimum radius to compute (cm)

    n_radii : int
        Number of radii to compute 
    """
    vrat = 2.2 
    pw = 1. / 3.
    f1 = ( 2.0*vrat / ( 1.0 + vrat) )**pw
    f2 = (( 2.0 / ( 1.0 + vrat ) )**pw) * (vrat**(pw-1.0))

    radius = r_min * vrat**(np.linspace(0,n_radii-1,n_radii)/3.)
    rup = f1*radius
    dr = f2*radius

    return radius, rup, dr


def picaso_format(opd, w0, g0 ):
    df = pd.DataFrame(index=[ i for i in range(opd.shape[0]*opd.shape[1])], columns=['lvl','w','opd','w0','g0'])
    i = 0 
    LVL = []
    WV,OPD,WW0,GG0 =[],[],[],[]
    for j in range(opd.shape[0]):
           for w in range(opd.shape[1]):
                LVL +=[j+1]
                WV+=[w+1]
                OPD+=[opd[j,w]]
                WW0+=[w0[j,w]]
                GG0+=[g0[j,w]]
    df.iloc[:,0 ] = LVL
    df.iloc[:,1 ] = WV
    df.iloc[:,2 ] = OPD
    df.iloc[:,3 ] = WW0
    df.iloc[:,4 ] = GG0
    return df

def available():
    """
    Print all available gas condensates 
    """
    pvs = [i for i in dir(pvaps) if i != 'np' and '_' not in i]
    gas_p = [i for i in dir(gas_properties) if i != 'np' and '_' not in i]
    return list(np.intersect1d(gas_p, pvs))

def recommend_gas(pressure, temperature, mh, mmw, plot=False, legend='inside',**plot_kwargs):
    """
    Recommends condensate species for a users calculation. 

    Parameters
    ----------
    pressure : ndarray, list
        Pressure grid for user's pressure-temperature profile. Unit=bars
    temperature : ndarray, list 
        Temperature grid (should be on same grid as pressure input). Unit=Kelvin
    mh : float 
        Metallicity in NOT log units. Solar =1 
    mmw : float 
        Mean molecular weight of the atmosphere. Solar = 2.2 
    plot : bool, optional
        Default is False. Plots condensation curves against PT profile to 
        demonstrate why it has chose the gases it did. 
    plot_kwargs : kwargs 
        Plotting kwargs for bokeh figure

    Returns
    -------
    ndarray, ndarray
        pressure (bars), condensation temperature (Kelvin)
    """    
    if plot: 
        from bokeh.plotting import figure, show
        from bokeh.models import Legend
        from bokeh.palettes import magma   
        plot_kwargs['y_range'] = plot_kwargs.get('y_range',[1e2,1e-3])
        plot_kwargs['plot_height'] = plot_kwargs.get('plot_height',400)
        plot_kwargs['plot_width'] = plot_kwargs.get('plot_width',600)
        plot_kwargs['x_axis_label'] = plot_kwargs.get('x_axis_label','Temperature (K)')
        plot_kwargs['y_axis_label'] = plot_kwargs.get('y_axis_label','Pressure (bars)')
        plot_kwargs['y_axis_type'] = plot_kwargs.get('y_axis_type','log')        
        fig = figure(**plot_kwargs)

    all_gases = available()
    cond_ts = []
    recommend = []
    line_widths = []
    for gas_name in all_gases: #case sensitive names
        #grab p,t from eddysed
        cond_p,t = condensation_t(gas_name, mh, mmw)
        cond_ts +=[t]

        interp_cond_t = np.interp(pressure,cond_p,t)

        diff_curve = interp_cond_t - temperature

        if ((len(diff_curve[diff_curve>0]) > 0) & (len(diff_curve[diff_curve<0]) > 0)):
            recommend += [gas_name]
            line_widths +=[5]
        else: 
            line_widths +=[1]        

    if plot: 
        legend_it = []
        ngas = len(all_gases)
        cols = magma(ngas)
        if legend is 'inside':
            fig.line(temperature,pressure, legend_label='User',color='black',line_width=5,line_dash='dashed')
            for i in range(ngas):

                fig.line(cond_ts[i],cond_p, legend_label=all_gases[i],color=cols[i],line_width=line_widths[i])
        else:
            f = fig.line(temperature,pressure, color='black',line_width=5,line_dash='dashed')
            legend_it.append(('input profile', [f]))
            for i in range(ngas):

                f = fig.line(cond_ts[i],cond_p ,color=cols[i],line_width=line_widths[i])
                legend_it.append((all_gases[i], [f]))

        if legend is 'outside':
            legend = Legend(items=legend_it, location=(0, 0))
            legend.click_policy="mute"
            fig.add_layout(legend, 'right')   
        
        plot_format(fig)
        show(fig) 

    return recommend 

def condensation_t(gas_name, mh, mmw, pressure =  np.logspace(-6, 2, 20)):
    """
    Find condensation curve for any planet given a pressure. These are computed 
    based on pressure vapor curves defined in pvaps.py. 

    Default is to compute condensation temperature on a pressure grid 

    Parameters
    ----------
    gas_name : str 
        Name of gas, which is case sensitive. See print_available to see which 
        gases are available. 
    mh : float 
        Metallicity in NOT log units. Solar =1 
    mmw : float 
        Mean molecular weight of the atmosphere. Solar = 2.2 
    pressure : ndarray, list, float, optional 
        Grid of pressures (bars) to compute condensation temperatures on. 
        Default = np.logspace(-3,2,20)

    Returns
    -------
    ndarray, ndarray
        pressure (bars), condensation temperature (Kelvin)
    """
    if isinstance(pressure,(float,int)):
        pressure = [pressure]
    temps = []
    for p in pressure: 
        temp = optimize.root_scalar(find_cond_t, 
                        bracket=[10, 10000], method='brentq', 
                        args=(p, mh, mmw, gas_name))
        temps += [temp.root]
    return np.array(pressure), np.array(temps)

def hot_jupiter():
    directory = os.path.join(os.path.dirname(__file__), "reference",
                                   "hj.pt")

    df = pd.read_csv(directory,delim_whitespace=True, usecols=[1,2,3],
                  names = ['pressure','temperature','kz'],skiprows=1)
    df.loc[df['pressure']>12.8,'temperature'] = np.linspace(1822,2100,df.loc[df['pressure']>12.8].shape[0])
    return df

def brown_dwarf(): 
    directory = os.path.join(os.path.dirname(__file__), "reference",
                                   "t1000g100nc_m0.0.dat")

    df  = pd.read_csv(directory,skiprows=1,delim_whitespace=True,
                 header=None, usecols=[1,2,3],
                 names=['pressure','temperature','chf'])
    return df




import numpy as np
from . import  pvaps
from . import  gas_properties
from scipy.stats import lognorm
from scipy.integrate import quad, simps
from scipy import optimize

def advdiff(qt, ad_qbelow=None,ad_qvs=None, ad_mixl=None,ad_dz=None ,ad_rainf=None,
        zb=None, b=None, eps=None, param='const'):
    """
    Calculate divergence from advective-diffusive balance for 
    condensate in a model layer

    All units are cgs
    
    A. Ackerman Feb-2000

    Parameters
    ----------
    qt : float 
        total mixing ratio of condensate + vapor (g/g)
    ad_qbelow : float 
        total mixing ratio of vapor in underlying layer (g/g)
    ad_qvs : float 
        saturation mixing ratio (g/g)
    ad_mixl : float 
        convective mixing length (cm)
    ad_dz : float 
        layer thickness (cm) 
    ad_rainf : float
        rain efficiency factor 
    zb : float
        altitude at bottom of layer
    b : float
        denominator of fsed exponential (if param is 'exp')
    param : str
        fsed parameterisation
        'const' (constant), 'exp' (exponential density derivation)

    Returns
    -------
    ad_qc : float 
        mixing ratio of condensed condensate (g/g)
    """
    #   All vapor in excess of saturation condenses
    if param is 'const':
        ad_qc = np.max([ 0., qt - ad_qvs ])

        # Eqn. 7 in A & M 
        #   Difference from advective-diffusive balance 
        advdif = ad_qbelow*np.exp( - ad_rainf*ad_qc*ad_dz / ( qt*ad_mixl ) )
        #print(advdif, ad_qc, ad_dz ,ad_mixl,qt )
    elif param is 'exp':
        fsed = ad_rainf; mixl = ad_mixl; z = ad_dz
        qc = (ad_qbelow - ad_qvs) * np.exp( - b * fsed / mixl * np.exp(zb/b) 
                            * (np.exp(z/b) -1) + eps*z/b)
        advdif = qc + ad_qvs

    advdif = advdif - qt
    return advdif



def vfall(r, grav,mw_atmos,mfp,visc,
              t,p, rhop):
    """
    Calculate fallspeed for a spherical particle at one layer in an
    atmosphere, depending on Reynolds number for Stokes flow.

    For Re_Stokes < 1, use Stokes velocity with slip correction
    For Re_Stokes > 1, use fit to Re = exp( b1*x + b2*x^2 )
     where x = log( Cd Re^2 / 24 )
     where b2 = -0.1 (curvature term) and 
     b1 from fit between Stokes at Re=1, Cd=24 and Re=1e3, Cd=0.45

    and Precipitation, Reidel, Holland, 1978) and Carlson, Rossow, and
    Orton (J. Atmos. Sci. 45, p. 2066, 1988)

    all units are cgs
    
    A. Ackerman Feb-2000

    Parameters
    ----------
    r : float
        particle radius (cm)
    grav : float 
        acceleration of gravity (cm/s^2)
    mw_atmos : float 
        atmospheric molecular weight (g/mol)
    mfp : float 
        atmospheric molecular mean free path (cm)
    visc : float 
        atmospheric dynamic viscosity (dyne s/cm^2) see Eqn. B2 in A&M
    t : float 
        atmospheric temperature (K)
    p  : float
        atmospheric pressure (dyne/cm^2)
    rhop : float 
        density of particle (g/cm^3)
    """

    # the drag coefficient for a reynolds number of 1000
    # which is appropriate for oblate spheroids 
    # Fig. 10-36 in Pruppacher & Klett 1978
    cdrag = 0.45 

    #In order to solve the drag problem we fit y=log(reynolds)
    #as a function of x=log(cdrag * reynolds**2)
    #if you assume that at reynolds= 1, cdrag=24 and 
    #reynolds=1000, cdrag=0.45 you get the following fit: 
    # y = 0.8 * x - 0.1 * x**2
    #Full explanation: see A & M Appendix B between eq. B2 and B3
    #Simply though, this allows us to get terminal fall velocity from 
    #reynolds number
    b1 = 0.8 
    b2 = -0.01 



    R_GAS = 8.3143e7 

    #calculate constants need to get Knudsen and Reynolds numbers
    knudsen = mfp / r
    rho_atmos = p / ( (R_GAS/mw_atmos) * t )
    drho = rhop - rho_atmos

    #Cunningham correction (slip factor for gas kinetic effects)
    #Cunningham, E., "On the velocity of steady fall of spherical particles through fluid medium," Proc. Roy. Soc. A 83(1910)357
    #Cunningham derived a value of 1.26 in the stone ages. In reality, this number is 
    #a function of the knudsen number. Various studies have derived 
    #different value for this number (see this citation
    #https://www.researchgate.net/publication/242470948_A_Novel_Slip_Correction_Factor_for_Spherical_Aerosol_Particles
    #Within the range of studied values, this 1.26 number changes particle sizes by a few microns
    #That is A OKAY for the level of accuracy we need. 
    beta_slip = 1. + 1.26*knudsen 

    #Stokes terminal velocity (low Reynolds number)
    #EQN B1 in A&M 
    #visc is eqn. B2 in A&M but is computed in `calc_qc`
    #also eqn 10-104 in Pruppacher & klett 1978
    vfall_r = beta_slip*(2.0/9.0)*drho*grav*r**2 / visc

    #compute reynolds number for low reynolds number case
    reynolds = 2.0*r*rho_atmos*vfall_r / visc

    #if reynolds number is between 1-1000 we are in turbulent flow 
    #limit
    if ((reynolds > 1) and (reynolds<=1e3)): #:#(reynolds >1e-2) and (reynolds <= 300)
        #OLD METHODLOGY
        #correct drag coefficient for turbulence (x = Cd Re^2 / 24)
        #x = np.log( reynolds )
        #y = b1*x + b2*x**2

        #compute cd * N_re^2 by equating drag and gravitational force  
        cd_nre2 = 32.0 * r**3.0 * drho * rho_atmos * grav / (3.0 * visc ** 2 ) 
        #coefficients from EQN 10-111 in Pruppachar & Klett 1978
        #they are an empirical fit to Figure 10-9
        xx = np.log(cd_nre2)
        b0,b1,b2,b3,b4,b5,b6 = -0.318657e1, 0.992696, -.153193e-2, -.987059e-3, -.578878e-3, 0.855176e-4, -0.327815e-5
        y = b0 + b1*xx**1 + b2*xx**2 + b3*xx**3 + b4*xx**4 + b5*xx**5 + b6*xx**6

        reynolds = np.exp(y)
        vfall_r = visc*reynolds / (2.*r*rho_atmos)

    if reynolds >1e3 :# 300
        #when Reynolds is greater than 1000, we can just use 
        #an asymptotic value that is independent of Reynolds number
        #Eqn. B3 from A&M 01
        vfall_r = beta_slip*np.sqrt( 8.*drho*r*grav / (3.*cdrag*rho_atmos) )

    return vfall_r 

def vfall_find_root(r, grav=None,mw_atmos=None,mfp=None,visc=None,
              t=None,p=None, rhop=None,w_convect=None ):
    """
    This is used to find F(X)-y=0 where F(X) is the fall speed for 
    a spherical particle and `y` is the convective velocity scale. 
    When the two are balanced we arrive at our correct particle radius.


    Therefore, it is the same as the `vfall` function but with the 
    subtraction of the convective velocity scale (cm/s) w_convect. 
    """
    vfall_r = vfall(r, grav,mw_atmos,mfp,visc,
              t,p, rhop )

    return vfall_r - w_convect

def force_balance(vf, r, grav, mw_atmos, mfp, visc, t, p, rhop, gas_kinetics=True):
    """"
    Define force balance for spherical particles falling in atmosphere, namely equate
    gravitational and viscous drag forces.

    Viscous drag assumed to be quadratic (Benchaita et al. 1983) and is a function of 
    the Reynolds number dependent drag coefficient.
    Drag coefficient taken from Khan-Richardson model (Richardson et al. 2002) and
    is valid for 1e-2 < Re < 1e5.

    Parameters
    ----------
    vf : float
        particle sedimentation velocity (cm/s)
    r : float
        particle radius (cm)
    grav : float 
        acceleration of gravity (cm/s^2)
    mw_atmos : float 
        atmospheric molecular weight (g/mol)
    mfp : float 
        atmospheric molecular mean free path (cm)
    visc : float 
        atmospheric dynamic viscosity (dyne s/cm^2) see Eqn. B2 in A&M
    t : float 
        atmospheric temperature (K)
    p  : float
        atmospheric pressure (dyne/cm^2)
    rhop : float 
        density of particle (g/cm^3)

    """

    R_GAS = 8.3143e7 
    rho_atmos = p / ( (R_GAS/mw_atmos) * t )
    # coefficients for drag coefficient taken from Khan-Richardson model (Richardson et al. 2002)
    # valid for 1e-2 < Re < 1e5
    a1 = 1.849; b1 = -0.31
    a2 = 0.293; b2 = 0.06
    b3 = 3.45    

    # include gas kinetic effects through slip factor 
    knudsen = mfp / r
    beta_slip = 1. + 1.26*knudsen 
    if gas_kinetics:
        vf = vf/beta_slip
                 
    # Reynolds number
    Re = rho_atmos * vf * 2 * r /  visc

    # Khan-Richardson approximation for drag coefficient is valid for 1e-2 < Re < 1e5
    RHS = (a1 * Re**b1 + a2 * Re**b2)**b3 * rho_atmos * np.pi * r**2 * vf**2 
       
    # gravitational force
    LHS = 4 * np.pi * r**3 * (rhop-rho_atmos) * grav / 3 

    return LHS - RHS

def solve_force_balance(solve_for, temp, grav, mw_atmos, mfp, visc, t, p, rhop, lo, hi):
    """
    This can be used to equate gravitational and viscous drag forces to find either
        (a) fallspeed for a spherical particle of given radius, or
        (b) the radius of a particle falling with a particular speed
    at one layer in an atmosphere.

    Parameters
    ----------
    solve_for : str
        property we are solving force balance for
        either "vfall" or "rw" 
    temp : float
        the other property needed to solve the force balance for solve_for
        ie if solve_for="vfall", temp will be a radius (cm)
           if solve_for="rw", temp will be w_convect (cm/s)
    grav : float 
        acceleration of gravity (cm/s^2)
    mw_atmos : float 
        atmospheric molecular weight (g/mol)
    mfp : float 
        atmospheric molecular mean free path (cm)
    visc : float 
        atmospheric dynamic viscosity (dyne s/cm^2) see Eqn. B2 in A&M
    t : float 
        atmospheric temperature (K)
    p  : float
        atmospheric pressure (dyne/cm^2)
    rhop : float 
        density of particle (g/cm^3)
    lo : float
        lower bound for root-finder
    hi : float
        upper bound for root-finder

    Returns
    ______
    soln.root : float
        the value of the parameter solve_for
    """

    def force_balance_new(u):
        if solve_for is "vfall":
            return force_balance(u, temp, grav, mw_atmos, mfp, visc, t, p, rhop) 
        elif solve_for is "rw":
            return force_balance(temp, u, grav, mw_atmos, mfp, visc, t, p, rhop) 

    soln = optimize.root_scalar(force_balance_new, bracket=[lo, hi], method='brentq') 

    return soln.root

def qvs_below_model(p_test, qv_dtdlnp=None, qv_p=None, 
    qv_t=None, qv_factor=None, qv_gas_name=None,mh=None,q_below=None):
    """
    Calculates the pressure of saturation mixing ratio for a gas 
    that is extrapolated below the model domain. 
    This is specifically used if the gas has saturated 
    below the model grid
    """
    #  Extrapolate temperature lapse rate to test pressure

    t_test = qv_t + np.log( qv_p / p_test )*qv_dtdlnp
    
    #  Compute saturation mixing ratio
    get_pvap = getattr(pvaps, qv_gas_name)
    if qv_gas_name == 'Mg2SiO4':
        pvap_test = get_pvap(t_test, p_test, mh=mh)
    else:
        pvap_test = get_pvap(t_test,mh=mh)

    fx = qv_factor * pvap_test / p_test 
    return np.log(fx) - np.log(q_below)


def find_cond_t(t_test, p_test = None, mh=None, mmw=None, gas_name=None):    
    """
    Root function used used to find condenstation temperature. E.g. 
    the temperature when  

    log p_vap = log partial pressure of gas 

    Parameters
    ----------
    t_test : float 
        Temp (K)
    p_test : float 
        Pressure bars 
    mh : float 
        NON log mh .. aka MH=1 for solar 
    mmw : float 
        mean molecular weight (2.2 for solar)
    gas_name : str 
        gas name, case sensitive 
    """
    pvap_fun = getattr(pvaps, gas_name)
    gas_p_fun = getattr(gas_properties, gas_name)
    #get gas mixing ratio 
    gas_mw, gas_mmr ,rho = gas_p_fun(mmw,mh=mh)
    #get vapor pressure and correct for masses of atmo and gas 
    if gas_name == 'Mg2SiO4':
        pv = gas_mw/mmw*pvap_fun(t_test,p_test, mh=mh)/1e6 #dynes to bars 
    else:
        pv = gas_mw/mmw*pvap_fun(t_test, mh=mh)/1e6 #dynes to bars 
    #get partial pressure
    partial_p = gas_mmr*p_test*mh 
    if pv == 0:
        pv = 1e-30 #dummy small
    return np.log10(pv) - np.log10(partial_p)

def moment(n, s, loc, scale, dist="lognormal"):
    """
    Calculate moment of size distribution.
    Will be extended to include more than just lognormal

    Parameters
    ----------
    n : float
        nth moment to be calculated
    s : float 
        std dev 
    loc : float
        Shift in distribution
    scale: float
        Scale distribution
    dist: str
        Continuous random variable for particle size distribution

    Returns
    -------
    moment_out : float 
        nth moment of distribution
    """

    #   for non-integer (any) n, must approximate moment integral numerically
    #   the following code works but it is slow 

    def pdf(r, s, loc, scale, dist):
        if dist=="lognormal":
            #return lognorm.pdf(r, s, loc, scale)
            return np.exp(- np.log((r-loc)/scale)**2 / (2*s**2)) / (r * s * np.sqrt(2. * np.pi))

    def func(r, n, s, loc, scale, dist):
        return r**n * pdf(r, s, loc, scale, dist)

    lbd = 0.
    ubd = np.inf
    moment_out = quad(func, lbd, ubd, epsabs=0, args=(n, s, loc, scale, dist))[0]

    #   for quicker result, take int(n) and use lognorm.moment, however, it will be slow for int(n) >= 2
    #n = int(round(n))
    #moment_out = lognorm.moment(n, s, loc, scale)
    return moment_out

def find_rg(rg, fsed, rw, alpha, s, loc=0., dist="lognormal"):
    """
    Root function used used to find the geometric mean radius of 
    lognormal size distribution.

    Useful if we consider more complicated size distributions for which
    analytical expressions for moments are not easily obtained.

    Parameters
    ----------
    rg : float 
        Geometric mean radius
    fsed : float 
        Sedimentation efficiency
    rw : float 
        Fall velocity particle radius 
    alpha : float 
        Exponent in power-law approximation for particle fall-speed
    s : float 
        s = log(sigma) where sigma is the geometric std dev of lognormal distn 
    loc : float
        Shift in lognormal distribution
    dist: str
        Continuous random variable for particle size distribution
    """

    return fsed - moment(3+alpha, s, loc, rg, dist) / rw**alpha / moment(3, s, loc, rg, dist)




import numpy as np
import pandas as pd
import astropy.units as u
import virga.justdoit as jdi
import virga.justplotit as jpi
import matplotlib.pyplot as plt
import time
from bokeh.plotting import show, figure
from virga.direct_mmr_solver import generate_altitude

#   locate data
mieff_directory = "~/Documents/codes/all-data/mieff_files"
fsed = 1
b = 2
eps = 10
Mark_data = False
refine_TP = False
quick_stop = False
generate = False

if Mark_data:
    TP_directory = "~/Documents/codes/all-data/Mark_data/"
    filenames = ["t1000g1000nc_m0.0.dat", "t1500g1000nc_m0.0.dat", "t1700g1000f3_m0.0k.dat", \
                    "t200g3160nc_m0.0.dat", "t2400g3160nc_m0.0.dat"]
    filename = TP_directory + filenames[1]
    
    #   define atmosphere properties
    df = pd.read_csv(filename, delim_whitespace=True, usecols=[1,2], header=None)
    df.columns = ["pressure", "temperature"]
    pressure = np.array(df["pressure"])[1:]
    temperature = np.array(df["temperature"])[1:]
    grav = df["pressure"][0] * 100 
    kz = 1e9
    
    metallicity = 1 #atmospheric metallicity relative to Solar
    mean_molecular_weight = 2.2 # atmospheric mean molecular weight
    #get pyeddy recommendation for which gases to run
    recommended_gases = jdi.recommend_gas(pressure, temperature,
                                                 metallicity, mean_molecular_weight)
    
    a = jdi.Atmosphere([recommended_gases[0]], fsed=fsed, mh=metallicity, mmw=mean_molecular_weight, b=b)
    a.gravity(gravity=grav, gravity_unit=u.Unit('cm/(s**2)'))
    a.ptk(df = pd.DataFrame({'pressure':pressure, 'temperature':temperature,
                                   'kz':kz}))

else:
    metallicity = 1 #atmospheric metallicity relative to Solar
    mean_molecular_weight = 2.2 # atmospheric mean molecular weight
    
    #set the run 
    #a = jdi.Atmosphere(['MnS','Cr','MgSiO3','Fe'],
    a = jdi.Atmosphere(['MnS'],#, 'Cr'],
                      fsed=fsed,mh=metallicity,
                     mmw = mean_molecular_weight, b=b)
    
    #set the planet gravity
    grav = 7.460
    a.gravity(gravity=grav, gravity_unit=u.Unit('m/(s**2)'))

    if generate:
        df = jdi.hot_jupiter()
        pres = np.array(df["pressure"])
        temp = np.array(df["temperature"])
        kz = np.array(df["kz"])
        gravity = grav*100
        print("initial number of pressure values = ", len(pres))

        plt.ylim(pres[len(pres)-1], pres[0])
        plt.loglog(temp, pres, label="initial")
        
        z, pres, P_z, temp, T_z, T_P, kz = generate_altitude(pres, temp, kz, gravity, 
                                                    mean_molecular_weight, refine_TP)  
        print("refined number of pressure values = ", len(pres))

        a.ptk(df = pd.DataFrame({'pressure':pres, 'temperature':temp,
                                   'kz':kz}))

        plt.loglog(temp, pres, "--", label="refined")
        plt.ylabel("pressure")
        plt.xlabel("temperature")
        plt.legend(loc="best")
        plt.savefig('temperature_profile.png')
        plt.show()

    else:
        #Get preset pt profile for testing
        a.ptk(df = jdi.hot_jupiter())

#   verify original and new solvers give same mixing ratios
labels = ["original", "new"]
lines = ["-", "--"]
solver = [True, False]
output = []
fig1, ax1 = plt.subplots()
for i in range(2):
    all_out = jdi.compute(a, as_dict=True, directory=mieff_directory)
    output.append(all_out)
    pres = all_out['pressure']
    qt = all_out['cond_plus_gas_mmr'][:,0]
    qc = all_out['condensate_mmr'][:,0]

    ax1.loglog(qt, pres, lines[i], label="qt " + labels[i] )
    ax1.loglog(qc, pres, lines[i], label="qc " + labels[i] )
    #ax1.loglog(qt-qc, pres, lines[i], label="qv " + labels[i] )

pres = output[0]["pressure"]
qc = output[0]["condensate_mmr"][:,0]
ax1.set_ylim(pres[len(pres)-1], pres[0])
#ax1.set_xlim([np.max([1e-9, np.min(qc*0.9)]), np.max(qc*1.1)])
ax1.set_ylabel("pressure")
ax1.legend(loc="best")
plt.savefig('mmr.png')
plt.show()

plt.ylim(pres[len(pres)-1], pres[0])
for i in range(2):
    pres_ = output[i]["pressure"]
    reff = output[i]["droplet_eff_r"][:,0]
    rg = output[i]["mean_particle_r"][:,0]
    plt.loglog(reff, pres_, lines[i], label="reff " + labels[i] )
    plt.loglog(rg, pres_, lines[i], label="rg " + labels[i] )
plt.ylabel("pressure")
plt.xlabel("radius")
plt.legend(loc="best")
plt.savefig('radii.png')
plt.show()

plt.ylim(pres[len(pres)-1], pres[0])
for i in range(2):
    pres_ = output[i]["pressure"]
    ndz = output[i]["column_density"][:,0]
    plt.loglog(ndz, pres_, lines[i], label="ndz " + labels[i] )
plt.ylabel("pressure")
plt.xlabel("column density")
plt.legend(loc="best")
plt.savefig('number_density.png')
plt.show()

import sys; sys.exit()

show(jpi.opd_by_gas(output[0]))
show(jpi.opd_by_gas(output[1]))
import sys; sys.exit()

z = all_out['altitude']
reff = all_out["droplet_eff_r"][:,0]
rg = all_out["mean_particle_r"][:,0]
ndz = all_out["column_density"][:,0]
opd = all_out["opd_per_layer"]
w0 = all_out["single_scattering"]
g0 = all_out["asymmetry"]
temp = all_out["temperature"]
reff = all_out["droplet_eff_r"]
rg = all_out["mean_particle_r"]
ndz = all_out["column_density"]
opd = all_out["opd_per_layer"]
w0 = all_out["single_scattering"]
g0 = all_out["asymmetry"]
    
#all_out = jdi.compute(a, as_dict=True, directory=mieff_directory)#, layers=True)
#pres_layer = all_out['pressure']
#qt_layer = all_out['cond_plus_gas_mmr'][:,0]
#qc_layer = all_out['condensate_mmr'][:,0]
#z_layer = all_out['altitude']
#temp_layer = all_out["temperature"]

plt.figure()
plt.ylim(pres[len(pres)-1], pres[0])
plt.loglog(qt, pres, '-', label="qt virga")
#plt.loglog(CR_qt, CR_pres, '--', label="qt caoimhe")
plt.loglog(qc, pres, '-', label="qc virga")
#plt.loglog(CR_qc, CR_pres, '--', label="qc caoimhe")
plt.loglog(qt-qc, pres, '-', label="qv virga")
#plt.loglog(CR_qt-CR_qc, CR_pres, '--', label="qv caoimhe")
plt.legend(loc="best")
plt.ylabel("pressure")
plt.show()
import sys; sys.exit()

plt.figure()
plt.loglog(reff, z, '-', label="effective radius")
plt.legend(loc="best")
plt.ylabel("altitude")
plt.show()

plt.figure()
plt.loglog(rg, z, '-', label="mean radius")
plt.legend(loc="best")
plt.ylabel("altitude")
plt.show()

plt.figure()
plt.loglog(ndz, z, '-', label="column density")
plt.legend(loc="best")
plt.ylabel("altitude")
plt.show()

plt.figure()
plt.loglog(opd, z, '-', label="optical density")
plt.legend(loc="best")
plt.ylabel("altitude")
plt.show()

plt.figure()
plt.loglog(w0, z, '-', label="single scattering")
plt.legend(loc="best")
plt.ylabel("altitude")
plt.show()

plt.figure()
plt.loglog(g0, z, '-', label="asymmetry")
plt.legend(loc="best")
plt.ylabel("altitude")
plt.show()



try:
    from pkg_resources import declare_namespace
    declare_namespace(__name__)
except ImportError:
    from pkgutil import extend_path
    __path__ = extend_path(__path__, __name__)


from . import justdoit 
from . import justplotit

__version__ = '0.3.2'


import numpy as np

def TiO2(temp, mh = 1 ):
	"""Computes vapor pressure curve
	
	Parameters 
	----------
	temp : float, ndarray 
		Temperature (K)
	mh : float 
		NON log metallicity relative to solar (1=1Xsolar)

	Returns
	-------
	vapor pressure in dyne/cm^2
	"""
	if mh != 1 : raise Exception("Warning: no M/H Dependence in vapor pressure curve for TiO2")

	return 1e6 * 10. ** (9.5489 - 32456.8678/temp)

def Cr(temp, mh = 1 ):
	"""Computes vapor pressure curve
	
	Parameters 
	----------
	temp : float, ndarray 
		Temperature (K)
	mh : float 
		NON log metallicity relative to solar (1=1Xsolar)

	Returns
	-------
	vapor pressure in dyne/cm^2
	"""
	if mh != 1 : raise Exception("Warning: no M/H Dependence in vapor pressure curve for Cr")
	mh = np.log10(mh)
	#Cr vapor pressure above cloud 
	pvap_cr_bars = 10.0**(7.2688-20353./temp)
	#Then convert from bars to dynes/cm^2    
	pvap_cr = pvap_cr_bars*1e6   
	return pvap_cr

def ZnS(temp,mh = 1 ):
	"""Computes vapor pressure curve
	
	Parameters 
	----------
	temp : float, ndarray 
		Temperature (K)
	mh : float 
		NON log metallicity relative to solar (1=1Xsolar)

	Returns
	-------
	vapor pressure in dyne/cm^2
	"""
	mh = np.log10(mh)
	#Zn vapor pressure above cloud 
	pvap_zns_bars = 10.0**(12.8117-15873./temp - mh)
	#Then convert from bars to dynes/cm^2    
	pvap_zns = pvap_zns_bars*1e6   
	return pvap_zns

def NH3(temp, mh = 1 ):
	"""Computes vapor pressure curve
	
	Parameters 
	----------
	temp : float, ndarray 
		Temperature (K)
	mh : float 
		NON log metallicity relative to solar (1=1Xsolar)

	Returns
	-------
	vapor pressure in dyne/cm^2
	"""
	if mh != 1 : raise Exception("Warning: no M/H Dependence in vapor pressure curve for NH3")
	mh = np.log10(mh)
	#NH3 vapor pressure above cloud 
	pvap_nh3 = np.exp(-86596./temp**2 - 2161./temp + 10.53)
	# convert from bars to dyne/cm^2
	pvap_nh3 = pvap_nh3*1e6    
	return pvap_nh3

def Na2S(temp,mh = 1 ):
	"""Computes vapor pressure curve
	
	Parameters 
	----------
	temp : float, ndarray 
		Temperature (K)
	mh : float 
		NON log metallicity relative to solar (1=1Xsolar)

	Returns
	-------
	vapor pressure in dyne/cm^2
	"""
	mh = np.log10(mh)
	#Na vapor pressure above cloud 
	#metallicityMH=0.0
	pvap_na2s_bars = 10.0**(8.5497-13889./temp-0.5*mh)
	#Then convert from bars to dynes/cm^2    
	pvap_na2s = pvap_na2s_bars*1e6 	
	return pvap_na2s

def MnS(temp, mh = 1 ):
	"""Computes vapor pressure curve
	
	Parameters 
	----------
	temp : float, ndarray 
		Temperature (K)
	mh : float 
		NON log metallicity relative to solar (1=1Xsolar)

	Returns
	-------
	vapor pressure in dyne/cm^2
	"""
	mh = np.log10(mh)
	#Mn vapor pressure above cloud 
	pvap_mns_bars = 10.0**(11.5315-23810./temp - mh)
	#Then convert from bars to dynes/cm^2    
	pvap_mns = pvap_mns_bars*1e6 
	return 	pvap_mns

def MgSiO3(temp, mh = 1 ):
	"""Computes vapor pressure curve
	
	Parameters 
	----------
	temp : float, ndarray 
		Temperature (K)
	mh : float 
		NON log metallicity relative to solar (1=1Xsolar)

	Returns
	-------
	vapor pressure in dyne/cm^2
	"""
	mh = np.log10(mh)
	#MgSiO3 vapor pressure above cloud
	#the one that is in A&M is this : np.exp(-58663./temp + 25.37)
	#this is a new one from Channon Visscher
	pvap_mgsio3 = 10.0**(11.83 - 27250.0/temp - mh)
	#convert bars -> dynes/cm^2
	pvap_mgsio3 = 1e6 * pvap_mgsio3 
	return pvap_mgsio3

def Mg2SiO4(temp, p, mh = 1 ):
	"""Computes vapor pressure curve
	
	Parameters 
	----------
	temp : float, ndarray 
		Temperature (K)
	mh : float 
		NON log metallicity relative to solar (1=1Xsolar)

	Returns
	-------
	vapor pressure in dyne/cm^2
	"""
	mh = np.log10(mh)
	#Another new expression from Channon Visscher, correspondance on 10/6/11
	#includes total pressure dependence and met dep. 
	pvap_mg2sio4 = 10.0**(-32488./temp + 14.88 - 0.2*np.log10(p/1e6) 
			- 1.4*mh) * 1e6 #convered from bars to dynes/cm2
	return pvap_mg2sio4


def KCl(temp, mh = 1 ):
	"""Computes vapor pressure curve
	
	Parameters 
	----------
	temp : float, ndarray 
		Temperature (K)
	mh : float 
		NON log metallicity relative to solar (1=1Xsolar)

	Returns
	-------
	vapor pressure in dyne/cm^2
	"""
	if mh != 1 : raise Exception("Warning: no M/H Dependence in vapor pressure curve for KCl")
	mh = np.log10(mh)
	pvap_kcl_bars = 10.0**(7.6106 - 11382./temp)
	#Then convert from bars to dynes/cm^2    
	pvap_kcl = pvap_kcl_bars*1e6  
	return pvap_kcl

def H2O(temp,do_buck = True,mh = 1 ):
	"""Computes vapor pressure curve
	
	Parameters 
	----------
	temp : float, ndarray 
		Temperature (K)
	mh : float 
		NON log metallicity relative to solar (1=1Xsolar)
	do_buck : bool 
		True means use Buck 1981 expresssion, False means use 
		Wexler's

	Returns
	-------
	vapor pressure in dyne/cm^2
	"""
	if mh != 1 : raise Exception("Warning: no M/H Dependence in vapor pressure curve for H2O")
	mh = np.log10(mh)
	if isinstance(temp, float): temp=np.array([temp]) 
	pvap_h2o = np.zeros(len(temp))
	#define constants used in Buck's expressions
	#Buck, 1981 (J. Atmos. Sci., 20, p. 1527)
	BAL = 6.1121e3 
	BBL = 18.729 
	BCL = 257.87 
	BDL = 227.3 
	BAI = 6.1115e3 
	BBI = 23.036 
	BCI = 279.82 
	BDI = 333.7 

	#define constants used in Wexler formulas
	#(see Flatau et al., 1992, J. Appl. Meteor. p. 1507)

	GG0 =-0.29912729e4
	GG1 =-0.60170128e4
	GG2 = 0.1887643854e2
	GG3 =-0.28354721e-1
	GG4 = 0.17838301e-4
	GG5 =-0.84150417e-9
	GG6 = 0.44412543e-12
	GG7 = 0.28584870e1

	HH0 = -0.58653696e4
	HH1 =  0.2224103300e2
	HH2 =  0.13749042e-1
	HH3 = -0.34031775e-4
	HH4 =  0.26967687e-7
	HH5 =  0.6918651

	t_low = np.where( temp < 273.16 )

	#Branch on temperature for liquid or ice
	if len(temp[t_low])>0:
		if do_buck:
			tc = temp - 273.16
			pvap_h2o[t_low] = BAI * np.exp( (BBI - tc/BDI)*tc / (tc + BCI) )
		else: 
			pvap_h2o[t_low] = 10*np.exp( 1.0/temp* 
						(HH0+(HH1+HH5*np.log(temp)+
						(HH2+(HH3+HH4*temp)*temp)*temp)*temp))

	#saturation vapor pressure over water
	t_med = np.where((temp>=273.16) & (temp<1048))
	if len(temp[t_med])>0:
		if do_buck: 
			tc = temp - 273.16
			pvap_h2o[t_med] = BAL * np.exp( (BBL - tc/BDL)*tc / (tc + BCL) )
		else: 
			pvap_h2o = 10*exp( (1.0/(temp*temp))* 
					( GG0+(GG1+(GG2+GG7*np.log(temp)+
					( GG3+(GG4+(GG5+GG6*temp)*temp)*temp)*temp)*temp)*temp ) )

	#anything greater than 1048 K is fixed at 600 bars
	t_high = np.where(temp>=1048)
	if len(temp[t_high])>0:
		pvap_h2o[t_high] = 600.0e6

	if len(pvap_h2o) == 1 : pvap_h2o = pvap_h2o[0]
	return pvap_h2o

def Fe(temp,mh = 1 ):
	"""Computes vapor pressure curve
	
	Parameters 
	----------
	temp : float, ndarray 
		Temperature (K)
	mh : float 
		NON log metallicity relative to solar (1=1Xsolar)

	Returns
	-------
	vapor pressure in dyne/cm^2
	"""
	if mh != 1 : raise Exception("Warning: no M/H Dependence in vapor pressure curve for Fe")
	mh = np.log10(mh)
	#EXPRESSION from Channon Visscher, correspondance on 6/3/11, added 7/27/11 (cvm)
	pvap_fe = 10.0**(7.09-20833./temp)
	pvap_fe = pvap_fe * 1e6   # convert from bars to dyne/cm^2
	return pvap_fe

def CH4(temp,mh = 1 ):
	"""Computes vapor pressure curve
	
	Parameters 
	----------
	temp : float, ndarray 
		Temperature (K)
	mh : float 
		NON log metallicity relative to solar (1=1Xsolar)

	Returns
	-------
	vapor pressure in dyne/cm^2
	"""
	if mh != 1 : raise Exception("Warning: no M/H Dependence in vapor pressure curve for CH4")
	mh = np.log10(mh)

	#	  AMR   -- molecular weight / ideal gas constant
	#     TCRIT -- triple point temperature
	#     PCRIT --    "     "   pressure
	#     AS    -- specific heat at constant pressure ( gas - solid )
	#     AL    --    "      "   "     "        "     ( gas - liquid )
	#     ALS   -- latent heat of sublimation
	#     ALV   --   "     "   "  vaporization


	AMR = 16.043 / 8.3143
	TCRIT = 90.68
	PCRIT = .11719 
	AS = 2.213 - 2.650
	AL = 2.213 - 3.370 
	ALS = 611.10
	ALV = 552.36 

	#ic=0: temperature below triple point
	#ic=1: temperature above triple point

	ic = 0
	if temp>TCRIT : ic = 1

	A, B, C = np.zeros(2),np.zeros(2),np.zeros(2)
	C[0] = - AMR * AS
	C[1] = - AMR * AL
	B[0] = - AMR * ( ALS + AS * TCRIT )
	B[1] = - AMR * ( ALV + AL * TCRIT )
	A[0] = PCRIT * TCRIT ** ( -C[0] ) * np.exp( -B[0] / TCRIT )
	A[1] = PCRIT * TCRIT ** ( -C[1] ) * np.exp( -B[1] / TCRIT )

	pvap_ch4 = A[ic] * temp**C[ic] * np.exp( B[ic] / temp )
	pvap_ch4= pvap_ch4*1e6    # convert from bars to dyne/cm^2

	return pvap_ch4

def Al2O3(temp,mh = 1 ):
	"""Computes vapor pressure curve
	
	Parameters 
	----------
	temp : float, ndarray 
		Temperature (K)
	mh : float 
		NON log metallicity relative to solar (1=1Xsolar)

	Returns
	-------
	vapor pressure in dyne/cm^2
	"""
	if mh != 1 : raise Exception("Warning: no M/H Dependence in vapor pressure curve for Al2O3")
	mh = np.log10(mh)
	#Kozasa et al. Ap J. 344 325
	#return np.exp(-73503./temp + 22.01)*1e6
	#calculated from wakeford 2017
	pvap_al2o3 = 1e6 * 10.0 ** (17.7 - 45892.6/temp - 1.66*mh)
	return pvap_al2o3




from __future__ import absolute_import, unicode_literals, print_function

import math, os

import time

import sys

import astropy.units as u

import virga.justdoit as jdi

#from bintools import binning

import pandas as pd

import dynesty

import concurrent.futures as cf

import pdb

import numpy as np

from numba import jit
import pickle


np.random.seed(0)

start_time=time.time()

########################

#VIRGA Setup

########################

metallicity = 1 #atmospheric metallicity relative to Solar
mean_molecular_weight = 2.2 # atmospheric mean molecular weight

# read in CARMA data
CARMA_files = '/Users/crooney/Documents/codes/all-data/CARMA/'

file = CARMA_files + 'pressure.csv'
df = pd.read_csv(file, sep=',',header=None)
pressure = df.values[0,:]/1e6

file = CARMA_files + 'temperature.csv'
df = pd.read_csv(file, sep=',',header=None)
temperature = df.values[0,:]

file = CARMA_files + 'kzz.csv'
df = pd.read_csv(file, sep=',',header=None)
kzz = df.values[0,:]

file = CARMA_files + 'opd.csv'
df = pd.read_csv(file, sep=',',header=None)
opd_carma = df.values[1:,:] # start at index 1 because carma gives opd at levels, virga gives opd at layers



# priors
def prior_transform(u):
    u[0] = 0.1 + 100*u[0] #alpha 
    u[1] = 0.1 + 5*u[1] #beta
    
    return u


#   log likelihood
def loglike(x):

    sum_planet = jdi.Atmosphere(['H2O'], fsed=x[0], mh=metallicity,
                         mmw = mean_molecular_weight,
                         b=x[1], param='exp')

    #set the planet gravity
    sum_planet.gravity(gravity=1e4, gravity_unit=u.Unit('cm/(s**2)'))

    #PT 
    sum_planet.ptk(df = pd.DataFrame({'pressure':pressure, 'temperature':temperature,
                           'kz':kzz}))

    #directory where mieff files are 
    mieff_dir = '/Users/crooney/Documents/codes/all-data/mieff_files'

    #get full dictionary output 
    all_out = sum_planet.compute(as_dict=True, directory=mieff_dir)

    avg_opd = np.mean(all_out['opd_per_layer'],1)
    y_mod = avg_opd

    loglikelihood=-0.5*np.sum((y_meas-y_mod)**2/err**2)
       
    return loglikelihood


y_meas = np.mean(opd_carma,1)
err = y_meas/100

########################

#MCMC output file name

########################

outname='VIRGA_RESULTS.pic'


#Dynesty Setup

########################

if __name__=='__main__':

    Nparam=2
    
    pool=cf.ProcessPoolExecutor()
    
    Nproc=2
    
    Nlive=1000
    
    dsampler = dynesty.NestedSampler(loglike, prior_transform, ndim=Nparam,
                    bound='multi', sample='auto', nlive=Nlive, update_interval=1.5, 
                    pool=pool, queue_size=Nproc)
    
    
    dsampler.run_nested(maxiter=9481)
    
    end_time=time.time()
    
    run_time=end_time-start_time
    
    pdf1=pd.DataFrame({'Time':[run_time]})
    
    pdf1.to_csv('runtime.csv',sep=' ',index=False)
    
    print('DONE')
    
    output=dsampler.results
    
    pickle.dump(output,open(outname,"wb"))





from scipy.integrate import solve_ivp
from scipy.interpolate import UnivariateSpline, interp1d
from scipy import optimize 
import numpy as np
import pandas as pd
from . import  pvaps
from .root_functions import vfall, vfall_find_root, find_rg, moment, solve_force_balance
import time
from . import justdoit as jdi

def direct_solver(temperature, pressure, condensibles, gas_mw, gas_mmr, rho_p , mw_atmos, 
                        gravity, kzz, fsed, mh, sig, rmin, nrad, d_molecule,eps_k,c_p_factor,
                        tol = 1e-15, refine_TP = True,og_vfall=True, analytical_rg = True):
    """
    Given an atmosphere and condensates, calculate size and concentration
    of condensates in balance between eddy diffusion and sedimentation.

    Parameters
    ----------
    temperature : ndarray
        Temperature at each layer (K)
    pressure : ndarray
        Pressure at each layer (dyn/cm^2)
    condensibles : ndarray or list of str
        List or array of condensible gas names
    gas_mw : ndarray
        Array of gas mean molecular weight from `gas_properties`
    gas_mmr : ndarray 
        Array of gas mixing ratio from `gas_properties`
    rho_p : float 
        density of condensed vapor (g/cm^3)
    mw_atmos : float 
        Mean molecular weight of the atmosphere
    gravity : float 
        Gravity of planet cgs
    kz : float or ndarray
        Kzz in cgs, either float or ndarray depending of whether or not 
        it is set as input
    fsed : float 
        Sedimentation efficiency, unitless
    mh : float 
        Atmospheric metallicity in NON log units (e.g. 1 for 1x solar)
    sig : float 
        Width of the log normal particle distribution
    d_molecule : float 
        diameter of atmospheric molecule (cm) (Rosner, 2000)
        (3.711e-8 for air, 3.798e-8 for N2, 2.827e-8 for H2)
        Set in Atmosphere constants 
    eps_k : float 
        Depth of the Lennard-Jones potential well for the atmosphere 
        Used in the viscocity calculation (units are K) (Rosner, 2000)
    c_p_factor : float 
        specific heat of atmosphere (erg/K/g) . Usually 7/2 for ideal gas
        diatomic molecules (e.g. H2, N2). Technically does slowly rise with 
        increasing temperature
    tol : float 
        Tolerance for direct solver
    refine_TP : bool
        Option to refine temperature-pressure profile for direct solver 
    analytical_rg : bool
        Option to use analytical expression for rg, or alternatively deduce rg from calculation
        Calculation option will be most useful for future inclusions of alternative particle size distributions

    Returns
    -------
    qc : ndarray 
        condenstate mixing ratio (g/g)
    qt : ndarray 
        gas + condensate mixing ratio (g/g)
    rg : ndarray
        geometric mean radius of condensate  (cm) 
    reff : ndarray
        droplet effective radius (second moment of size distrib, cm)
    ndz : ndarray 
        number column density of condensate (cm^-3)
    qc_path : ndarray 
        vertical path of condensate 
    pres : ndarray
        Pressure at each layer (dyn/cm^2)
    temp : ndarray
        Temperature at each layer (K)
    z : ndarray
        altitude of each layer (cm)
    
    """

    ngas =  len(condensibles)
    # refine temperature-pressure profile
    # this improves the accuracy of the mmr calculation
    (z, pres, P_z, temp, T_z, T_P, kz) = generate_altitude(pressure, temperature, kzz, gravity, 
                                                                mw_atmos, refine_TP) 

    pres_out = pressure
    temp_out = temperature
    z_out = interp1d(pres, z)(pres_out)

    mixl_out = np.zeros((len(pres_out), ngas))
    qc_out = np.zeros((len(pres_out), ngas))
    qt_out = np.zeros((len(pres_out), ngas))
    rg_out = np.zeros((len(pres_out), ngas))
    reff_out = np.zeros((len(pres_out), ngas))
    ndz_out = np.zeros((len(pres_out), ngas))
    qc_path = np.zeros(ngas)

    # find mmr and particle distribution for every condensible
    # perform calculation on refined TP profile but output values corresponding to initial profile
    for i, igas in zip(range(ngas), condensibles):
        gas_name = igas
        qc, qt, rg, reff, ndz, dz, qc_path[i], mixl = calc_qc(z, P_z, T_z, T_P, kz,
            gravity, gas_name, gas_mw[i], gas_mmr[i], rho_p[i], mw_atmos, mh, fsed, sig, rmin, nrad, 
            d_molecule,eps_k,c_p_factor,
            tol,og_vfall, analytical_rg)

        # generate qc values for original pressure data
        qc_out[:,i] = interp1d(pres, qc)(pres_out)
        qt_out[:,i] = interp1d(pres, qt)(pres_out)
        rg_out[:,i] = interp1d(pres, rg)(pres_out)
        reff_out[:,i] = interp1d(pres, reff)(pres_out)
        mixl_out[:,i] = interp1d(pres, mixl)(pres_out)

        ndz_temp = ndz/dz
        dz_new = np.insert(-(z_out[1:]-z_out[:-1]), len(z_out)-1, 1e-8)
        ndz_out[:,i] = interp1d(pres, ndz_temp)(pres_out) * dz_new

    return (qc_out, qt_out, rg_out, reff_out, ndz_out, qc_path, pres_out, temp_out, z_out,mixl_out)

def calc_qc(z, P_z, T_z, T_P, kz, gravity, gas_name, gas_mw, gas_mmr, rho_p, mw_atmos, 
                    mh, fsed, sig, rmin, nrad, d_molecule,eps_k,c_p_factor,
                    tol, og_vfall=True, analytical_rg=True, supsat=0):
    """
    Calculate condensate optical depth and effective radius for atmosphere,
    assuming geometric scatterers. 

    z : float 
        Altitude  cm 
    P_z: function
        Pressure at altitude z (dyne/cm^2)
    T_z: function
        Temperature at altitude z (K)
    T_P: function
        Temperature at pressure P (K)
    kz : float or ndarray
        Kzz in cgs
    gravity : float 
        Gravity of planet cgs 
    gas_name : str 
        Name of condensate 
    gas_mw : ndarray
        Array of gas mean molecular weight from `gas_properties`
    gas_mmr : ndarray 
        Array of gas mixing ratio from `gas_properties`
    rho_p : float 
        density of condensed vapor (g/cm^3)
    mw_atmos : float 
        Mean molecular weight of the atmosphere
    mh : float 
        Metallicity NON log solar (1 = 1x solar)
    fsed : float 
        Sedimentation efficiency (unitless)
    sig : float 
        Width of the log normal particle distrubtion 
    d_molecule : float 
        diameter of atmospheric molecule (cm) (Rosner, 2000)
        (3.711e-8 for air, 3.798e-8 for N2, 2.827e-8 for H2)
        Set in Atmosphere constants 
    eps_k : float 
        Depth of the Lennard-Jones potential well for the atmosphere 
        Used in the viscocity calculation (units are K) (Rosner, 2000)
    c_p_factor : float 
        specific heat of atmosphere (erg/K/g) . Usually 7/2 for ideal gas
        diatomic molecules (e.g. H2, N2). Technically does slowly rise with 
        increasing temperature
    tol : float 
        Tolerance for direct solver
    analytical_rg : bool
        Option to use analytical expression for rg, or alternatively deduce rg from calculation
        Calculation option will be most useful for future inclusions of alternative particle size distributions
    supsat : float, optional
        Default = 0 , Saturation factor (after condensation)

    Returns
    -------
    qc_out : ndarray 
        condenstate mixing ratio (g/g)
    qt_out : ndarray 
        gas + condensate mixing ratio (g/g)
    rg : ndarray
        geometric mean radius of condensate  (cm) 
    reff : ndarray
        droplet effective radius (second moment of size distrib, cm)
    ndz : ndarray 
        number column density of condensate (cm^-3)
    qc_path : ndarray 
        vertical path of condensate 
    """
    #   universal gas constant (erg/mol/K)
    R_GAS = 8.3143e7
    AVOGADRO = 6.02e23
    K_BOLTZ = R_GAS / AVOGADRO
    PI = np.pi 
    #   specific gas constant for atmosphere (erg/K/g)
    r_atmos = R_GAS / mw_atmos
    #   specific gas constant for cloud (erg/K/g)
    r_cloud = R_GAS/ gas_mw
    #   atmospheric mean free path (cm)
    def mfp(T, P):
        #   atmospheric number density (molecules/cm^3)
        n_atmos = P / ( K_BOLTZ * T )
        return  1. / ( np.sqrt(2.) * n_atmos * PI * d_molecule**2 )
    #   atmospheric scale height (cm) 
    def scale_h(T): 
        return r_atmos * T / gravity
    #   atmospheric density (g/cm^3)
    def rho_atmos(T, P):
        return P / ( r_atmos * T )
    def dHdP(P):
        return r_atmos * dTdP(P) / gravity
    #   lapse ratio ** not sure what this physically is
    def lapse_ratio(T, P):
        dTdP =  T_P.derivative()
        def dTdlnP(P):
            return P * dTdP(P)
        return dTdlnP(P) / ( T / c_p_factor )
    #   convective mixing length scale (cm) 
    def mixl(T, P):
        return np.max( [0.10, lapse_ratio(T, P)]) * scale_h(T)
    get_pvap = getattr(pvaps, gas_name)
    def pvap(T, P):
        if gas_name == 'Mg2SiO4':
            return get_pvap(float(T), float(P), mh=mh)
        else:
            return get_pvap(float(T), mh=mh)
    #   mass mixing ratio of saturated layer
    def qvs(T, P):
        qvs_val =  (supsat + 1) * pvap(T, P) / (r_cloud * T) / rho_atmos(T, P) 
        return qvs_val
    # atmospheric viscosity (dyne s/cm^2)
    # EQN B2 in A & M 2001, originally from Rosner+2000
    # Rosner, D. E. 2000, Transport Processes in Chemically Reacting Flow Systems (Dover: Mineola)
    def visc(T):
        return (5./16.*np.sqrt( PI * K_BOLTZ * T * (mw_atmos/AVOGADRO)) /
        ( PI * d_molecule**2 ) /
        ( 1.22 * ( T / eps_k )**(-0.16) ))
    #   convective velocity scale (cm/s) from mixing length theory
    def w_convect(T, P, kz):
        return kz / mixl(T, P) 

    ##  Define and solve ODE (4) in AM2001 using scipy solve_ivp ---------------------------------------------
    q_below = gas_mmr
    z = z[::-1]
    kz = kz[::-1]
    #   define ode
    def AM4(z, q):
        P = P_z(z); T = T_z(z); 
        qc_val = max([0., q - qvs(T, P)])
        return - fsed *  qc_val / mixl(T, P)

    sol = solve_ivp(lambda t, y: AM4(t, y), [z[0], z[len(z)-1]], [q_below], method = "RK23", 
            rtol = 1e-12, atol = tol, dense_output=True, t_eval=z)
    qt = sol.sol

    mixl_out = np.zeros(len(z))
    qc_out = np.zeros(len(z))
    qt_out = np.zeros(len(z))
    p_out = np.zeros(len(z))
    for i in range(len(z)):
        p_out[i] = P_z(z[i])
        qt_out[i] = qt(z[i])
        T = T_z(z[i]); P = P_z(z[i])
        qc_out[i] = max([0., qt_out[i] - qvs(T, P)])
        mixl_out[i] =  mixl(T, P)
    #   --------------------------------------------------------------------
    #   Find <rw> corresponding to <w_convect> using function vfall()

    #   precision of vfall solution (cm/s)
    dz = np.insert(z[1:]-z[:-1], 0, 1e-8)
    rw = np.zeros(len(qc_out))
    rg = np.zeros(len(qc_out))
    reff = np.zeros(len(qc_out))
    ndz = np.zeros(len(qc_out))
    qc_path = 0.
    for i in range(len(qc_out)):

        if qc_out[i] == 0.0: # layer is cloud free
            rg[i] = 0.; reff[i] = 0; ndz[i] = 0.

        else:
            #   range of particle radii to search (cm)
            rlo = 1.e-10
            rhi = 10.
            find_root = True
            while find_root:
                try:
                    P = p_out[i]; T = T_P(p_out[i]); k = kz[i]
                    if og_vfall:
                        rw_temp = optimize.root_scalar(vfall_find_root, bracket=[rlo, rhi], method='brentq', 
                            args=(gravity, mw_atmos, mfp(T, P),  visc(T), T, P, rho_p, w_convect(T, P, k)))
                    else:
                        rw_temp = solve_force_balance("rw", w_convect(T, P, k), gravity, mw_atmos, mfp(T, P),
                                                    visc(T), T, P, rho_p, rlo, rhi)
                    find_root = False
                except ValueError:
                    rlo = rlo/10
                    rhi = rhi*10

            #fall velocity particle radius 
            if og_vfall: rw[i] = rw_temp.root
            else: rw[i] = rw_temp
    
            #   geometric std dev of lognormal size distribution ** sig is the geometric std dev
            lnsig2 = 0.5*np.log( sig )**2
            #   sigma floor for the purpose of alpha calculation
            sig_alpha = np.max( [1.1, sig] )    

            #if fsed > 1 :
            #    #   Bulk of precip at r > rw: exponent between rw and rw*sig
            #    r_ = rw[i]*sig_alpha
            #else:
            #    #   Bulk of precip at r < rw: exponent between rw/sig and rw
            #    r_ = rw[i]/sig_alpha

            #if og_vfall:
            #    vf = vfall(r_, gravity, mw_atmos, mfp(T,P), visc(T), T, P,  rho_p)
            #else:
            #    vlo = 1e0; vhi = 1e6
            #    vf = solve_force_balance("vfall", r_, gravity, mw_atmos, mfp(T, P),
            #                                        visc(T), T, P, rho_p, vlo, vhi)

            #alpha = (np.log( vf / w_convect(T, P, k) )
            #                     / np.log( r_ / rw[i] ))

            #   find alpha for power law fit vf = w(r/rw)^alpha
            def pow_law(r, alpha):
                return np.log(w_convect(T, P, k)) + alpha * np.log (r / rw[i]) 

            r_, rup, dr = jdi.get_r_grid(r_min = rmin, n_radii = nrad)
            vfall_temp = []
            for j in range(len(r_)):
                if og_vfall:
                    vfall_temp.append(vfall(r_[j], gravity, mw_atmos, mfp(T, P), visc(T), T, P, rho_p))
                else:
                    vlo = 1e0; vhi = 1e6
                    find_root = True
                    while find_root:
                        try:
                            vfall_temp.append(solve_force_balance("vfall", r_[j], gravity, mw_atmos, 
                                mfp(T, P), visc(T), T, P, rho_p, vlo, vhi))
                            find_root = False
                        except ValueError:
                            vlo = vlo/10
                            vhi = vhi*10

            pars, cov = optimize.curve_fit(f=pow_law, xdata=r_, ydata=np.log(vfall_temp), p0=[0], 
                                bounds=(-np.inf, np.inf))
            alpha = pars[0]

            if analytical_rg:
                #     EQN. 13 A&M 
                #   geometric mean radius of lognormal size distribution
                rg[i] = (fsed**(1./alpha) *
                        rw[i] * np.exp(-(alpha + 6) * lnsig2))

                #   droplet effective radius (cm)
                reff[i] = rg[i] * np.exp(5 * lnsig2)

                #      EQN. 14 A&M
                #   column droplet number concentration (cm^-2)
                ndz[i] = (3 * rho_atmos(T, P) * qc_out[i] * dz[i] /
                            ( 4 * np.pi * rho_p * rg[i]**3 ) * np.exp(-9 * lnsig2))

            else:
                #   range of particle radii to search (cm)
                rlo = 1.e-10
                rhi = 1.e2
                #   geometric mean radius of size distribution
                rg_temp = optimize.root_scalar(find_rg, bracket=[rlo, rhi], method='brentq', 
                                                    args=(fsed, rw[i], alpha, np.log(sig)))
                rg[i] = rg_temp.root

                #   droplet effective radius (cm)
                #   ratio of third to second moment of size distribution
                reff[i] = moment(3, np.log(sig), 0., rg[i]) / moment(2, np.log(sig), 0., rg[i])

                #   column droplet number concentration (cm^-2)
                ndz[i] = (3 * fsed * rw[i]**alpha * qc_out[i] * rho_atmos(T, P) * dz[i] / 
                            (4 * np.pi * rho_p * moment(3+alpha, np.log(sig), 0., rg[i])))


        if i > 0:   
            qc_path = (qc_path + qc_out[i-1] *
                            ( p_out[i-1] - p_out[i] ) / gravity)

    return (qc_out[::-1], qt_out[::-1], rg[::-1], reff[::-1], ndz[::-1], dz[::-1], qc_path, mixl_out[::-1])

def generate_altitude(pres, temp, kz, gravity, mw_atmos, refine_TP=True, eps=10):
    """
    Refine temperature pressure profile according to maximum temperature-difference
    between pressure layers.

    pres : ndarray
        Pressure at each layer (dyn/cm^2)
    temp : ndarray
        Temperature at each layer (K)
    kz : float or ndarray
        Kzz in cgs
    gravity : float 
        Gravity of planet cgs 
    mw_atmos : float 
        Mean molecular weight of the atmosphere
    refine_TP : bool
        Option to refine temperature-pressure profile for direct solver 
    eps : float
        maximum temperature difference between pressure layers

    Returns
    -------
    z : float 
        Altitude  cm 
    pres_ : ndarray
        Pressure at each layer (dyn/cm^2)
    P_z: function
        Pressure at altitude z (dyne/cm^2)
    temp_ : ndarray
        Temperature at each layer (K)
    T_z: function
        Temperature at altitude z (K)
    T_P: function
        Temperature at pressure P (K)
    kz_ : float or ndarray
        Kzz in cgs
    """
    #   universal gas constant (erg/mol/K)
    R_GAS = 8.3143e7
    #   specific gas constant for atmosphere (erg/K/g)
    r_atmos = R_GAS / mw_atmos
    #   atmospheric scale height (cm) 
    def H(T): 
        return r_atmos * T / gravity

    T_P = UnivariateSpline(pres, temp)
    # this is grim fix this
    if len(pres) == len(kz):
        kz_P = interp1d(pres, kz) 
    else:
        kz_P = interp1d(pres, kz[:-1]) 

    pres_ = pres
    if refine_TP:
        #   we use barometric formula which assumes constant temperature 
        #   define maximum difference between temperature values which if exceeded, reduce pressure stepsize
        n = len(pres_)
        while max(abs(T_P(pres_[1:]) - T_P(pres_[:-1]))) > eps:
            indx = np.where(abs(T_P(pres_[1:]) - T_P(pres_[:-1])) > eps)[0]
            mids = pres_[indx] + (pres_[indx+1] - pres_[indx]) / 2
            pres_ = np.insert(pres_, indx+1, mids)
    pres_ = pres_[::-1]
    
    z = np.zeros(len(pres_))
    T = np.zeros(len(pres_)); T[0] = temp[len(temp)-1]
    K = np.zeros(len(pres_)); K[0] = kz[len(kz)-1]
    for i in range(len(pres_) - 1):
        T[i+1] = T_P(pres_[i+1])
        K[i+1] = kz_P(pres_[i+1])
        dz = - H(T[i+1]) * np.log(pres_[i+1] / pres_[i]) 
        z[i+1] = z[i] + dz
            
    P_z = UnivariateSpline(z, pres_)
    T_z = UnivariateSpline(z, T)
    temp_ = T
    kz_ = K
    
    return (z[::-1], pres_[::-1], P_z, temp_[::-1], T_z, T_P, kz_[::-1])
    #return (z, pres_, P_z, temp_, T_z, T_P, kz_)
